# -*- coding: utf-8 -*-
from bda.plone.orders.tests import Orders_INTEGRATION_TESTING
from plone.testing import layered
from plone.testing.zca import UNIT_TESTING

import doctest
import pprint  # noqa
import unittest


optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS

TESTFILES = [
    ("dynamicmailtemplate.rst", UNIT_TESTING),
    ("dynamicmaillibrary.rst", Orders_INTEGRATION_TESTING),
]


def test_suite():
    return unittest.TestSuite(
        [
            layered(
                doctest.DocFileSuite(
                    filename, optionflags=optionflags, globs={"pprint": pprint.pprint}
                ),
                layer=layer,
            )
            for filename, layer in TESTFILES
        ]
    )



# -*- coding: utf-8 -*-
from .events import IBookingCancelledEvent  # noqa: W291,F401
from .events import IBookingReservedToOrderedEvent  # noqa: W291,F401
from .events import IOrderSuccessfulEvent  # noqa: W291,F401
from .events import IStockThresholdReached  # noqa: W291,F401
from .mailtemplates import IDynamicMailTemplateLibrary  # noqa: W291,F401
from .mailtemplates import IDynamicMailTemplateLibraryStorage  # noqa: W291,F401
from .markers import IBuyable  # noqa: W291,F401
from .markers import IOrdersExtensionLayer  # noqa: W291,F401
from .markers import IVendor  # noqa: W291,F401
from .notifications import IGlobalNotificationText  # noqa: W291,F401
from .notifications import IItemNotificationText  # noqa: W291,F401
from .notifications import INotificationSettings  # noqa: W291,F401
from .notifications import IPaymentText  # noqa: W291,F401
from .orders import IBookingData  # noqa: W291,F401
from .orders import IInvoiceSender  # noqa: W291,F401
from .orders import IOrderData  # noqa: W291,F401
from .orders import IOrderState  # noqa: W291,F401
from .tradings import ITrading  # noqaL W291,F401
from .workflow import SALARIED_FAILED  # noqa: W291,F401
from .workflow import SALARIED_MIXED  # noqa: W291,F401
from .workflow import SALARIED_NO  # noqa: W291,F401
from .workflow import SALARIED_TRANSITION_OUTSTANDING  # noqa: W291,F401
from .workflow import SALARIED_TRANSITION_SALARIED  # noqa: W291,F401
from .workflow import SALARIED_YES  # noqa: W291,F401
from .workflow import STATE_CANCELLED  # noqa: W291,F401
from .workflow import STATE_FINISHED  # noqa: W291,F401
from .workflow import STATE_MIXED  # noqa: W291,F401
from .workflow import STATE_NEW  # noqa: W291,F401
from .workflow import STATE_PROCESSING  # noqa: W291,F401
from .workflow import STATE_RESERVED  # noqa: W291,F401
from .workflow import STATE_TRANSITION_CANCEL  # noqa: W291,F401
from .workflow import STATE_TRANSITION_FINISH  # noqa: W291,F401
from .workflow import STATE_TRANSITION_PROCESS  # noqa: W291,F401
from .workflow import STATE_TRANSITION_RENEW  # noqa: W291,F401



# -*- coding: utf-8 -*-
from bda.plone.orders import message_factory as _
from bda.plone.orders.interfaces import IDynamicMailTemplateLibraryStorage
from bda.plone.orders.mailtemplates import DEFAULT_TEMPLATE_ATTRS
from bda.plone.orders.mailtemplates import DynamicMailTemplate
from bda.plone.orders.mailtemplates import REQUIRED_TEMPLATE_ATTRS
from Products.Five import BrowserView
from yafowil.base import ExtractionError
from yafowil.plone.form import YAMLForm


TEMPLATE = DynamicMailTemplate(
    required=REQUIRED_TEMPLATE_ATTRS, defaults=DEFAULT_TEMPLATE_ATTRS
)


class MailtemplatesView(BrowserView):
    def default_attrs(self):
        items = []
        for key in sorted(DEFAULT_TEMPLATE_ATTRS.keys()):
            items.append(
                {
                    "placeholder": key.replace(".", "_"),
                    "example": DEFAULT_TEMPLATE_ATTRS[key],
                }
            )
        return items

    def rendered(self):
        tpllib = IDynamicMailTemplateLibraryStorage(self.context)
        items = []
        for key in tpllib.direct_keys():
            preview = TEMPLATE(tpllib[key], DEFAULT_TEMPLATE_ATTRS)
            items.append({"title": key, "preview": preview})
        return items


class MailtemplatesForm(YAMLForm):
    """Form to edit all the mailtemplates
    """

    form_template = "bda.plone.orders.browser:forms/mailtemplates.yaml"
    message_factory = _

    def value_tpl(self, widget, data):
        tpllib = IDynamicMailTemplateLibraryStorage(self.context)
        value = []
        for key in tpllib.direct_keys():
            value.append({"title": key, "template": tpllib[key]})
        return value

    def validate_tpl(self, widget, data):
        state, msg = TEMPLATE.validate(data.extracted)
        if not state:
            raise ExtractionError(msg)
        return data.extracted

    def form_action(self, widget, data):
        return "%s/@@mailtemplates" % self.context.absolute_url()

    def save(self, widget, data):
        tpllib = IDynamicMailTemplateLibraryStorage(self.context)
        newkeys = []
        for record in data.extracted["array"]:
            newkeys.append(record["title"])
            tpllib[record["title"]] = record["template"]
        for key in tpllib.direct_keys():
            if key not in newkeys:
                del tpllib[key]
        self.request.response.redirect(self.form_action(widget, data))

    def next(self, request):
        return True



# -*- coding: utf-8 -*-
from bda.plone.orders.datamanagers.base import BaseState
from bda.plone.orders.datamanagers.base import calculate_order_salaried
from bda.plone.orders.datamanagers.base import calculate_order_state
from bda.plone.orders.datamanagers.order import OrderData
from bda.plone.orders.interfaces.orders import IBookingData
from repoze.catalog.query import Any
from repoze.catalog.query import Eq
from zope.interface import implementer

import uuid


def booking_update_comment(context, booking_uid, comment):
    booking_data = BookingData(context, uid=booking_uid)
    if booking_data.booking is None:
        raise ValueError("invalid value (booking)")
    booking = booking_data.booking
    booking.attrs["buyable_comment"] = comment
    booking_data.reindex()


@implementer(IBookingData)
class BookingData(BaseState):
    def __init__(self, context, uid=None, booking=None, vendor_uids=[]):
        """Create booking data object by criteria

        :param context: Context to work with
        :type object: Plone or Content instance
        :param uid: Booking uid. XOR with booking
        :type uid: string or uuid.UUID object
        :param booking: Booking record. XOR with uid
        :type booking: souper.soup.Record object
        :param vendor_uids: Vendor uids, used to filter order bookings.
        :type vendor_uids: List of vendor uids as string or uuid.UUID object.
        """
        if not (bool(uid) != bool(booking)):  # ^= xor
            raise ValueError("uid and booing are mutually exclusive")
        if uid and not isinstance(uid, uuid.UUID):
            uid = uuid.UUID(uid)
        vendor_uids = [uuid.UUID(str(vuid)) for vuid in vendor_uids]
        self.context = context
        self._uid = uid
        self._booking = booking

        self.vendor_uids = vendor_uids

    @property
    def uid(self):
        if self._uid:
            return self._uid
        return self.booking.attrs["uid"]

    def reindex(self):
        return

    @property
    def booking(self):
        if self._booking:
            return self._booking
        soup = self.bookings_soup
        query = Eq("uid", self.uid)
        if self.vendor_uids:
            query = query & Any("vendor_uid", self.vendor_uids)
        result = soup.query(query, with_size=True)
        length = next(result)
        if length != 1:  # first result is length
            return None
        self._booking = next(result)
        return self._booking

    @property
    def order(self):
        # XXX: rename to order_data for less confusion
        return OrderData(
            self.context,
            uid=self.booking.attrs["order_uid"],
            vendor_uids=self.vendor_uids,
        )

    @property
    def state(self):
        return self.booking.attrs["state"]

    @state.setter
    def state(self, value):
        # XXX: currently we need to delete attributes before setting to a new
        #      value in order to persist change. fix in appropriate place.
        booking = self.booking
        self.update_item_stock(booking, booking.attrs["state"], value)
        del booking.attrs["state"]
        booking.attrs["state"] = value
        order = self.order
        del order.order.attrs["state"]
        order.order.attrs["state"] = calculate_order_state(order.bookings)
        self.reindex_bookings([booking])
        self.reindex_order(order.order)

    @property
    def salaried(self):
        return self.booking.attrs["salaried"]

    @salaried.setter
    def salaried(self, value):
        # XXX: currently we need to delete attributes before setting to a new
        #      value in order to persist change. fix in appropriate place.
        booking = self.booking
        del booking.attrs["salaried"]
        booking.attrs["salaried"] = value
        order = self.order
        del order.order.attrs["salaried"]
        order.order.attrs["salaried"] = calculate_order_salaried(order.bookings)  # noqa
        self.reindex_bookings([booking])
        self.reindex_order(order.order)



# -*- coding: utf-8 -*-
STATE_CANCELLED = "cancelled"
STATE_FINISHED = "finished"
STATE_MIXED = "mixed"
STATE_NEW = "new"
STATE_PROCESSING = "processing"
STATE_RESERVED = "reserved"

STATE_TRANSITION_CANCEL = "cancel"
STATE_TRANSITION_FINISH = "finish"
STATE_TRANSITION_PROCESS = "process"
STATE_TRANSITION_RENEW = "renew"

SALARIED_YES = "yes"
SALARIED_NO = "no"
SALARIED_MIXED = "mixed"
SALARIED_FAILED = "failed"

SALARIED_TRANSITION_SALARIED = "mark_salaried"
SALARIED_TRANSITION_OUTSTANDING = "mark_outstanding"



# -*- coding: utf-8 -*-
from bda.plone.cart.interfaces import IShippingItem
from bda.plone.cart.utils import get_object_by_uid
from bda.plone.orders import message_factory as _
from bda.plone.orders.common import acquire_vendor_or_shop_root
from bda.plone.orders.common import get_bookings_soup
from bda.plone.orders.common import get_order
from bda.plone.orders.common import get_orders_soup
from bda.plone.orders.contacts import extract_contact
from bda.plone.orders.contacts import get_contacts_soup
from bda.plone.orders.contacts import lookup_contact
from bda.plone.orders.contacts import LOOKUP_QUERY_MAPPING
from bda.plone.orders.datamanagers.base import calculate_order_salaried
from bda.plone.orders.datamanagers.base import calculate_order_state
from bda.plone.orders.datamanagers.order import OrderData
from bda.plone.orders.interfaces import ITrading
from bda.plone.payment import Payments
from decimal import Decimal
from node.ext.zodb.utils import reset_odict
from plone.uuid.interfaces import IUUID
from repoze.catalog.query import Eq
from zope.component.hooks import getSite

import logging
import uuid


logger = logging.getLogger("bda.plone.orders UPGRADE")


def fix_bookings_vendor_uid(ctx=None):
    """Add vendor_uid attribute to booking records.
    """
    portal = getSite()
    soup = get_bookings_soup(portal)
    data = soup.storage.data
    need_rebuild = False
    for item in data.values():
        update = False
        try:
            item.attrs["vendor_uid"]
            if not isinstance(item.attrs["vendor_uid"], uuid.UUID):
                update = True
        except KeyError:
            update = True
        if not update:
            continue
        buyable_uid = item.attrs["buyable_uid"]
        obj = get_object_by_uid(portal, buyable_uid)
        if not obj:
            shop = acquire_vendor_or_shop_root(portal)
        else:
            shop = acquire_vendor_or_shop_root(obj)
        vendor_uid = uuid.UUID(IUUID(shop))
        item.attrs["vendor_uid"] = vendor_uid
        need_rebuild = True
        logging.info(u"Added vendor_uid to booking {0}".format(item.attrs["uid"]))
    if need_rebuild:
        soup.rebuild()
        logging.info("Rebuilt bookings catalog")


def fix_orders_vendor_uids(ctx=None):
    """Add vendor_uids attribute to order records.
    """
    portal = getSite()
    soup = get_orders_soup(portal)
    data = soup.storage.data
    need_rebuild = False
    for item in data.values():
        update = False
        try:
            item.attrs["vendor_uids"]
            if (
                not isinstance(item.attrs["vendor_uids"], list)
                or not item.attrs["vendor_uids"]
            ):
                update = True
        except KeyError:
            update = True
        if not update:
            continue
        order_data = OrderData(portal, order=item)
        vendor_uids = set()
        for booking in order_data.bookings:
            vendor_uids.add(booking.attrs["vendor_uid"])
        item.attrs["vendor_uids"] = list(vendor_uids)
        need_rebuild = True
        logging.info(u"Added vendor_uids to order {0}".format(item.attrs["uid"]))
    if need_rebuild:
        soup.rebuild()
        logging.info("Rebuilt orders catalog")


def fix_bookings_state_salaried_tid(ctx=None):
    portal = getSite()
    soup = get_orders_soup(portal)
    data = soup.storage.data
    need_rebuild = False
    for item in data.values():
        order_data = OrderData(portal, order=item)
        try:
            state = item.attrs["state"]
            state_exists = True
        except KeyError:
            state = None
            state_exists = False
        try:
            salaried = item.attrs["salaried"]
            salaried_exists = True
        except KeyError:
            salaried = None
            salaried_exists = False
        try:
            tid = item.attrs["tid"]
            tid_exists = True
        except KeyError:
            tid = "none"  # tid default in b.p.payment
            tid_exists = False
        for booking in order_data.bookings:
            # add too booking node
            try:
                booking.attrs["state"]
            except KeyError:
                booking.attrs["state"] = state
                need_rebuild = True
                logging.info(
                    u"Added state {0} to booking {1}".format(state, item.attrs["uid"])
                )
            try:
                booking.attrs["salaried"]
            except KeyError:
                booking.attrs["salaried"] = salaried
                need_rebuild = True
                logging.info(
                    u"Added salaried {0} to booking {1}".format(
                        salaried, item.attrs["uid"]
                    )
                )
            try:
                booking.attrs["tid"]
            except KeyError:
                booking.attrs["tid"] = tid
                need_rebuild = True
                logging.info(
                    u"Added tid {0} to booking {1}".format(tid, item.attrs["uid"])
                )
        # now, delete from order node
        if state_exists:
            del item.attrs["state"]
        if salaried_exists:
            del item.attrs["salaried"]
        if tid_exists:
            del item.attrs["tid"]
    if need_rebuild:
        bookings_soup = get_bookings_soup(portal)
        bookings_soup.rebuild()
        logging.info("Rebuilt bookings catalog")


def fix_discount_attrs(ctx=None):
    portal = getSite()
    # discount attrs on order
    orders_soup = get_orders_soup(portal)
    need_rebuild = False
    data = orders_soup.storage.data
    for item in data.values():
        try:
            item.attrs["cart_discount_net"]
        except KeyError:
            need_rebuild = True
            item.attrs["cart_discount_net"] = Decimal(0)
            logging.info(
                u"Added cart_discount_net to order {0}".format(item.attrs["uid"])
            )
        try:
            item.attrs["cart_discount_vat"]
        except KeyError:
            need_rebuild = True
            item.attrs["cart_discount_vat"] = Decimal(0)
            logging.info(
                u"Added cart_discount_vat to order {0}".format(item.attrs["uid"])
            )
    if need_rebuild:
        orders_soup.rebuild()
        logging.info("Rebuilt orders catalog")
    # discount attrs on bookings
    bookings_soup = get_bookings_soup(portal)
    need_rebuild = False
    data = bookings_soup.storage.data
    for item in data.values():
        try:
            item.attrs["discount_net"]
        except KeyError:
            need_rebuild = True
            item.attrs["discount_net"] = Decimal(0)
            logging.info(u"Added discount_net to booking {0}".format(item.attrs["uid"]))
    if need_rebuild:
        bookings_soup.rebuild()
        logging.info("Rebuilt bookings catalog")


def fix_shipping_attrs(ctx=None):
    portal = getSite()
    orders_soup = get_orders_soup(portal)
    data = orders_soup.storage.data
    for item in data.values():
        try:
            item.attrs["shipping_method"]
        except KeyError:
            item.attrs["shipping_method"] = "unknown"
            logging.info(
                u"Added shipping_method {0} to booking {1}".format(
                    "unknown", item.attrs["uid"]
                )
            )
        try:
            item.attrs["shipping_label"]
        except KeyError:
            item.attrs["shipping_label"] = _("unknown", default=u"Unknown")
            logging.info(
                u"Added shipping_label {0} to booking {1}".format(
                    "unknown", item.attrs["uid"]
                )
            )
        try:
            item.attrs["shipping_description"]
        except KeyError:
            item.attrs["shipping_description"] = _("unknown", default=u"Unknown")
            logging.info(
                u"Added shipping_description {0} to booking {1}".format(
                    "unknown", item.attrs["uid"]
                )
            )
        try:
            item.attrs["shipping_net"]
        except KeyError:
            item.attrs["shipping_net"] = item.attrs["shipping"]
            logging.info(
                u"Added shipping_net {0} to booking {1}".format(
                    item.attrs["shipping"], item.attrs["uid"]
                )
            )
        try:
            item.attrs["shipping_vat"]
        except KeyError:
            item.attrs["shipping_vat"] = Decimal(0)
            logging.info(
                u"Added shipping_vat {0} to booking {1}".format(
                    Decimal(0), item.attrs["uid"]
                )
            )


def fix_payment_attrs(ctx=None):
    portal = getSite()
    payments = Payments(portal)
    orders_soup = get_orders_soup(portal)
    data = orders_soup.storage.data
    for item in data.values():
        try:
            item.attrs["payment_method"]
            item.attrs["payment_label"]
            continue
        except KeyError:
            payment_method = item.attrs["payment_selection.payment"]
            payment = payments.get(payment_method)
            if payment:
                payment_label = payment.label
            else:
                payment_label = _("unknown", default=u"Unknown")
            item.attrs["payment_method"] = payment_method
            logging.info(
                u"Added payment_method {0} to booking {1}".format(
                    payment_method, item.attrs["uid"]
                )
            )
            item.attrs["payment_label"] = payment_label
            logging.info(
                u"Added payment_label {0} to booking {1}".format(
                    payment_label, item.attrs["uid"]
                )
            )


def fix_bookings_shippable(ctx=None):
    portal = getSite()
    soup = get_bookings_soup(portal)
    data = soup.storage.data
    need_rebuild = False
    for booking in data.values():
        try:
            booking.attrs["shippable"]
        except KeyError:
            obj = get_object_by_uid(portal, booking.attrs["buyable_uid"])
            shippable = True
            if obj:
                shippable = IShippingItem(obj).shippable
            booking.attrs["shippable"] = shippable
            need_rebuild = True
            logging.info(
                u"Added shippable {0} to booking {1}".format(
                    shippable, booking.attrs["uid"]
                )
            )
    if need_rebuild:
        bookings_soup = get_bookings_soup(portal)
        bookings_soup.rebuild()
        logging.info("Rebuilt bookings catalog")


def fix_bookings_trading(ctx=None):
    portal = getSite()
    soup = get_bookings_soup(portal)
    data = soup.storage.data
    need_rebuild = False
    for booking in data.values():
        try:
            booking.attrs["item_number"]
        except KeyError:
            obj = get_object_by_uid(portal, booking.attrs["buyable_uid"])
            if obj:
                trading = ITrading(obj)
                item_number = trading.item_number
                gtin = trading.gtin
            else:
                item_number = ""
                gtin = ""
            need_rebuild = True
            booking.attrs["item_number"] = item_number
            logging.info(
                u"Added item_number {0} to booking {1}".format(
                    item_number, booking.attrs["uid"]
                )
            )
            booking.attrs["gtin"] = gtin
            logging.info(
                u"Added gtin {0} to booking {1}".format(gtin, booking.attrs["uid"])
            )
    if need_rebuild:
        bookings_soup = get_bookings_soup(portal)
        bookings_soup.rebuild()
        logging.info("Rebuilt bookings catalog")


def reset_records(ctx=None):
    def ignore_key(key):
        return key.startswith("____")

    portal = getSite()
    soup = get_orders_soup(portal)
    data = soup.storage.data
    for order in data.values():
        reset_odict(order.attrs.storage, ignore_key=ignore_key)
        logging.info(
            u"Reset attributes storage on order {0}".format(order.attrs["uid"])
        )
    soup = get_bookings_soup(portal)
    data = soup.storage.data
    for booking in data.values():
        reset_odict(booking.attrs.storage, ignore_key=ignore_key)
        logging.info(
            u"Reset attributes storage on booking {0}".format(booking.attrs["uid"])
        )


def fix_bookings_email(ctx=None):
    """Add email attribute to booking records from the corresponding order.
    """
    portal = getSite()
    soup = get_bookings_soup(portal)
    data = soup.storage.data
    need_rebuild = False
    for item in data.values():
        update = False
        try:
            item.attrs["email"]
        except KeyError:
            update = True
        if not update:
            continue

        order = get_order(portal, item.attrs["order_uid"])
        email = order.attrs.get("personal_data.email", "n/a")

        item.attrs["email"] = email
        need_rebuild = True
        logging.info(u"Added email to booking {0}".format(item.attrs["uid"]))
    if need_rebuild:
        soup.rebuild()
        logging.info("Rebuilt bookings catalog")


def fix_contacts_email(ctx=None):
    """Add email attribute to contact records.
    """
    portal = getSite()
    soup = get_contacts_soup(portal)
    data = soup.storage.data
    need_rebuild = False
    for item in data.values():
        update = False
        try:
            item.attrs["email"]
        except KeyError:
            update = True
        if not update:
            continue

        email = item.attrs.get("personal_data.email", "n/a")
        item.attrs["email"] = email
        need_rebuild = True
        logging.info(u"Added email to contact {0}".format(item.attrs["uid"]))
    if need_rebuild:
        soup.rebuild()
        logging.info("Rebuilt contacts catalog")


def fix_order_state_and_salaried(ctx=None):
    """Re-add state and salaried on order, needed for sorting in orders table
    """
    portal = getSite()
    soup = get_orders_soup(portal)
    data = soup.storage.data
    for order in data.values():
        order_data = OrderData(portal, uid=order.attrs["uid"])
        bookings = order_data.bookings
        order.attrs["state"] = calculate_order_state(bookings)
        order.attrs["salaried"] = calculate_order_salaried(bookings)
    soup.rebuild()


def fix_order_contact_uid(ctx=None):
    """Add contact_uid to order
    """
    portal = getSite()
    soup = get_orders_soup(portal)
    data = soup.storage.data
    need_rebuild = False
    for order in data.values():
        if not order.attrs.get("contact_uid", None):
            contact = lookup_contact(portal, extract_contact(order))
            if contact:
                order.attrs["contact_uid"] = contact.attrs["uid"]
            else:
                order.attrs["contact_uid"] = ""
            need_rebuild = True
            logging.info(u"Added contact_uid to order {0}".format(order.attrs["uid"]))
    if need_rebuild:
        soup.rebuild()
        logging.info("Rebuilt contacts catalog")



# -*- coding: utf-8 -*-
from bda.plone.orders.interfaces import IBookingCancelledEvent
from bda.plone.orders.interfaces import IBookingReservedToOrderedEvent
from bda.plone.orders.interfaces import IOrderSuccessfulEvent
from bda.plone.orders.interfaces import IStockThresholdReached
from zope.interface import implementer


@implementer(IOrderSuccessfulEvent)
class OrderSuccessfulEvent(object):
    def __init__(self, context, request, order_uid, booking_attrs=None):
        self.context = context
        self.request = request
        self.order_uid = order_uid
        self.booking_attrs = booking_attrs


@implementer(IBookingCancelledEvent)
class BookingCancelledEvent(object):
    def __init__(self, context, request, order_uid, booking_attrs):
        self.context = context
        self.request = request
        self.order_uid = order_uid
        self.booking_attrs = booking_attrs


@implementer(IBookingReservedToOrderedEvent)
class BookingReservedToOrderedEvent(object):
    def __init__(self, context, request, order_uid, booking_attrs):
        self.context = context
        self.request = request
        self.order_uid = order_uid
        self.booking_attrs = booking_attrs


@implementer(IStockThresholdReached)
class StockThresholdReached(object):
    def __init__(self, context, request, order_uid, stock_threshold_reached_items):
        self.context = context
        self.request = request
        self.order_uid = order_uid
        self.stock_threshold_reached_items = stock_threshold_reached_items



# -*- coding: utf-8 -*-
from bda.plone.orders import mailnotify as MN

import unittest


class TestMailnotifyUnit(unittest.TestCase):
    def test_indent_wrap(self):
        """The _indent mehtod should wrap like defined by it's parameters.
        """
        txt = u"abcd " * 3
        ctrl = "     abcd\nabcd abcd"  # textwrap removes whitespace at EOL
        res = MN._indent(txt, width=10, ind=5)
        self.assertEqual(res, ctrl)

    def test_indent_unicode(self):
        """The _indent method should be able to handle non-ASCII data.
        """
        txt = "äüöß " * 3
        ctrl = u"äüöß äüöß\näüöß"  # textwrap removes whitespace at EOL
        res = MN._indent(txt, width=10, ind=0)
        self.assertEqual(res, ctrl)



# -*- coding: utf-8 -*-
from zope.interface import Interface


class IDynamicMailTemplateLibrary(Interface):
    """A set of named templates.
    """

    def keys():
        """list names of templates.
        """

    def __getitem__(name):
        """return template by name.
        """


class IDynamicMailTemplateLibraryStorage(IDynamicMailTemplateLibrary):
    def direct_keys():
        """non acquired keys.
        """

    def __setitem__(name, template):
        """store template under a name.
        """

    def __delitem__(name):
        """remove template with this name.
        """



# -*- coding: utf-8 -*-
from AccessControl import Unauthorized
from bda.plone.orders import interfaces as ifaces
from bda.plone.orders import message_factory as _
from bda.plone.orders import permissions
from bda.plone.orders import vocabularies as vocabs
from bda.plone.orders.browser.common import customers_form_vocab
from bda.plone.orders.browser.common import salaried_form_vocab
from bda.plone.orders.browser.common import states_form_vocab
from bda.plone.orders.browser.common import Transition
from bda.plone.orders.browser.common import Translate
from bda.plone.orders.browser.common import vendors_form_vocab
from bda.plone.orders.browser.dropdown import BaseDropdown
from bda.plone.orders.common import DT_FORMAT
from bda.plone.orders.common import get_vendor_by_uid
from bda.plone.orders.common import get_vendor_uids_for
from bda.plone.orders.common import get_vendors_for
from bda.plone.orders.datamanagers.order import OrderData
from bda.plone.orders.interfaces import IBuyable
from bda.plone.orders.transitions import do_transition_for
from bda.plone.orders.transitions import transitions_of_main_state
from bda.plone.orders.transitions import transitions_of_salaried_state
from Products.CMFPlone.interfaces import IPloneSiteRoot
from Products.CMFPlone.resources import add_bundle_on_request
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from repoze.catalog.query import Any
from repoze.catalog.query import Contains
from repoze.catalog.query import Eq
from souper.soup import get_soup
from souper.soup import LazyRecord
from yafowil.base import factory
from yafowil.utils import Tag
from zope.i18n import translate

import json
import plone.api
import six.moves.urllib.parse
import uuid


###############################################################################
# orders table base
###############################################################################


class TableData(BrowserView):
    """Base view for displaying sour records in datatable.
    """

    soup_name = None
    search_text_index = None

    @property
    def columns(self):
        """Return list of dicts with column definitions:

        [{
            'id': 'colid',
            'label': 'Col Label',
            'head': callback,
            'renderer': callback,
        }]
        """
        raise NotImplementedError(
            u"Abstract DataTable does not implement " u"``columns``."
        )

    def query(self, soup):
        """Return 2-tuple with result length and lazy record iterator.
        """
        raise NotImplementedError(
            u"Abstract DataTable does not implement " u"``query``."
        )

    def sort(self):
        columns = self.columns
        sortparams = dict()
        sortcols_idx = int(self.request.form.get("iSortCol_0"))
        sortparams["index"] = columns[sortcols_idx]["id"]
        sortparams["reverse"] = self.request.form.get("sSortDir_0") == "desc"
        return sortparams

    def all(self, soup):
        data = soup.storage.data
        sort = self.sort()
        sort_index = soup.catalog[sort["index"]]
        iids = sort_index.sort(list(data.keys()), reverse=sort["reverse"])

        def lazyrecords():
            for iid in iids:
                yield LazyRecord(iid, soup)

        return soup.storage.length.value, lazyrecords()

    def slice(self, fullresult):
        start = int(self.request.form["iDisplayStart"])
        length = int(self.request.form["iDisplayLength"])
        count = 0
        for lr in fullresult:
            if count >= start and count < (start + length):
                yield lr
            if count >= (start + length):
                break
            count += 1

    def column_def(self, colname):
        for column in self.columns:
            if column["id"] == colname:
                return column

    def __call__(self):
        soup = get_soup(self.soup_name, self.context)
        aaData = list()
        length, lazydata = self.query(soup)
        columns = self.columns
        colnames = [_["id"] for _ in columns]
        # XXX: include JSON response header

        def record2list(record):
            result = list()
            for colname in colnames:
                coldef = self.column_def(colname)
                renderer = coldef.get("renderer")
                if renderer:
                    value = renderer(colname, record)
                else:
                    value = record.attrs.get(colname, "")
                result.append(value)
            return result

        for lazyrecord in self.slice(lazydata):
            aaData.append(record2list(lazyrecord()))
        data = {
            "sEcho": int(self.request.form["sEcho"]),
            "iTotalRecords": soup.storage.length.value,
            "iTotalDisplayRecords": length,
            "aaData": aaData,
        }
        self.request.response.setHeader("Content-type", "application/json")
        return json.dumps(data)


###############################################################################
# order related dropdowns and transitions
###############################################################################


class OrderDropdown(BaseDropdown):
    @property
    def order_data(self):
        vendor_uid = self.request.form.get("vendor", "")
        if vendor_uid:
            vendor_uids = [vendor_uid]
        else:
            vendor_uids = get_vendor_uids_for()
        return OrderData(self.context, order=self.record, vendor_uids=vendor_uids)


class OrderStateDropdown(OrderDropdown):
    name = "state"
    css = "dropdown change_order_state_dropdown"
    action = "orderstatetransition"
    vocab = vocabs.state_vocab()
    transitions = vocabs.state_transitions_vocab()

    @property
    def value(self):
        return self.order_data.state

    @property
    def items(self):
        transitions = transitions_of_main_state(self.value)
        return self.create_items(transitions)


class OrderSalariedDropdown(OrderDropdown):
    name = "salaried"
    css = "dropdown change_order_salaried_dropdown"
    action = "ordersalariedtransition"
    vocab = vocabs.salaried_vocab()
    transitions = vocabs.salaried_transitions_vocab()

    @property
    def value(self):
        return self.order_data.salaried or ifaces.SALARIED_NO

    @property
    def items(self):
        transitions = transitions_of_salaried_state(self.value)
        return self.create_items(transitions)


class OrderTransition(Transition):
    def do_transition(self, uid, transition, vendor_uids):
        order_data = OrderData(self.context, uid=uid, vendor_uids=vendor_uids)
        do_transition_for(
            order_data,
            transition=transition,
            context=self.context,
            request=self.request,
        )
        return order_data.order


class OrderStateTransition(OrderTransition):
    dropdown = OrderStateDropdown


class OrderSalariedTransition(OrderTransition):
    dropdown = OrderSalariedDropdown


class OrdersViewBase(BrowserView):
    table_view_name = "@@orderstable"

    def __init__(self, context, request):
        super(OrdersViewBase, self).__init__(context, request)
        add_bundle_on_request(request, "bdajax-jquerytools")
        add_bundle_on_request(request, "bdajax-jquerytools-overlay")
        add_bundle_on_request(request, "datatables")
        add_bundle_on_request(request, "bda-plone-orders")

    def orders_table(self):
        return self.context.restrictedTraverse(self.table_view_name)()


###############################################################################
# orders views
###############################################################################


class OrdersView(OrdersViewBase):
    def __call__(self):
        # check if authenticated user is vendor
        if not get_vendors_for():
            raise Unauthorized
        return super(OrdersView, self).__call__()


class MyOrdersView(OrdersViewBase):
    table_view_name = "@@myorderstable"


class OrdersTableBase(BrowserView):
    table_template = ViewPageTemplateFile("templates/table.pt")
    table_id = "bdaploneorders"
    data_view_name = "@@ordersdata"

    def rendered_table(self):
        return self.table_template(self)

    def render_filter(self):
        return None

    def render_order_actions_head(self):
        return None

    def render_order_actions(self, colname, record):
        return None

    def render_salaried(self, colname, record):
        salaried = OrderData(self.context, order=record).salaried or ifaces.SALARIED_NO
        return translate(vocabs.salaried_vocab()[salaried], context=self.request)

    def render_state(self, colname, record):
        state = OrderData(self.context, order=record).state
        if not state:
            return "-/-"
        return translate(vocabs.state_vocab()[state], context=self.request)

    def render_dt(self, colname, record):
        value = record.attrs.get(colname, "")
        if value:
            value = value.strftime(DT_FORMAT)
        return value

    @property
    def ajaxurl(self):
        return u"{0}/{1}".format(self.context.absolute_url(), self.data_view_name)

    @property
    def columns(self):
        return [
            {
                "id": "actions",
                "label": _("actions", default=u"Actions"),
                "head": self.render_order_actions_head,
                "renderer": self.render_order_actions,
            },
            {
                "id": "created",
                "label": _("date", default=u"Date"),
                "renderer": self.render_dt,
            },
            {
                "id": "personal_data.lastname",
                "label": _("lastname", default=u"Last Name"),
            },
            {
                "id": "personal_data.firstname",
                "label": _("firstname", default=u"First Name"),
            },
            {"id": "personal_data.email", "label": _("email", default=u"Email")},
            {"id": "billing_address.city", "label": _("city", default=u"City")},
            {
                "id": "salaried",
                "label": _("salaried", default=u"Salaried"),
                "renderer": self.render_salaried,
            },
            {
                "id": "state",
                "label": _("state", default=u"State"),
                "renderer": self.render_state,
            },
        ]


class OrdersTable(OrdersTableBase):
    def render_filter(self):
        # vendor areas of current user
        vendors = vendors_form_vocab()
        vendor_selector = None
        # vendor selection, include if more than one vendor
        if len(vendors) > 2:
            vendor_selector = factory(
                "div:label:select",
                name="vendor",
                value=self.request.form.get("vendor", ""),
                props={
                    "vocabulary": vendors,
                    "label": _("filter_for_vendors", default=u"Filter for vendors"),
                },
            )
        # customers of current user
        customers = customers_form_vocab()
        customer_selector = None
        # customers selection, include if more than one customer
        if len(customers) > 2:
            customer_selector = factory(
                "div:label:select",
                name="customer",
                value=self.request.form.get("customer", ""),
                props={
                    "vocabulary": customers,
                    "label": _("filter_for_customers", default=u"Filter for customers"),
                },
            )

        states = states_form_vocab()
        state_selector = factory(
            "div:label:select",
            name="state",
            value=self.request.form.get("state", ""),
            props={
                "vocabulary": states,
                "label": _("filter_for_state", default=u"Filter for states"),
            },
        )

        salaried = salaried_form_vocab()
        salaried_selector = factory(
            "div:label:select",
            name="salaried",
            value=self.request.form.get("salaried", ""),
            props={
                "vocabulary": salaried,
                "label": _("filter_for_salaried", default=u"Filter for salaried state"),
            },
        )

        # concatenate filters
        filter_widgets = ""
        if vendor_selector:
            filter_widgets += vendor_selector(request=self.request)
        if customer_selector:
            filter_widgets += customer_selector(request=self.request)

        filter_widgets += state_selector(request=self.request)
        filter_widgets += salaried_selector(request=self.request)

        return filter_widgets

    def render_order_actions_head(self):
        tag = Tag(Translate(self.request))
        select_all_orders_attrs = {
            "name": "select_all_orders",
            "type": "checkbox",
            "class_": "select_all_orders",
            "title": _("select_all_orders", default=u"Select all visible orders"),
        }
        select_all_orders = tag("input", **select_all_orders_attrs)
        notify_customers_target = self.context.absolute_url()
        notify_customers_attributes = {
            "ajax:target": notify_customers_target,
            "class_": "notify_customers",
            "href": "",
            "title": _(
                "notify_customers", default=u"Notify customers of selected orders"
            ),
        }
        notify_customers = tag("a", "&nbsp;", **notify_customers_attributes)
        return select_all_orders + notify_customers

    def render_order_actions(self, colname, record):
        tag = Tag(Translate(self.request))
        vendor_uid = self.request.form.get("vendor", "")
        base_url = self.context.absolute_url()

        # view order
        if vendor_uid:
            view_order_target = "%s?uid=%s&vendor=%s" % (
                base_url,
                str(record.attrs["uid"]),
                vendor_uid,
            )
        else:
            view_order_target = "%s?uid=%s" % (base_url, str(record.attrs["uid"]))
        view_order_attrs = {
            "ajax:bind": "click",
            "ajax:target": view_order_target,
            "ajax:overlay": "order",
            "href": "",
            "title": _("view_order", default=u"View Order"),
        }
        order_icon_url = "%s/++resource++bda.plone.orders/order.png" % base_url
        view_order_icon = tag("img", src=order_icon_url)
        view_order = tag("a", view_order_icon, **view_order_attrs)

        # view invoice
        view_invoice_target = "%s?uid=%s" % (base_url, str(record.attrs["uid"]))
        view_invoice_attrs = {
            "ajax:bind": "click",
            "ajax:target": view_invoice_target,
            "ajax:overlay": "invoice",
            "ajax:overlay-css": "invoice_overlay",
            "href": "",
            "title": _("view_invoice", default=u"View Invoice"),
        }
        invoice_icon_url = "%s/++resource++bda.plone.orders/invoice.png" % base_url
        view_invoice_icon = tag("img", src=invoice_icon_url)
        view_invoice = tag("a", view_invoice_icon, **view_invoice_attrs)

        # select order
        select_order_attrs = {
            "name": "select_order",
            "type": "checkbox",
            "value": record.attrs["uid"],
            "class_": "select_order",
        }
        select_order = tag("input", **select_order_attrs)

        # return joined actions
        return select_order + view_order + view_invoice

    def check_modify_order(self, order):
        vendor_uid = self.request.form.get("vendor", "")
        if vendor_uid:
            vendor_uids = [vendor_uid]
            vendor = get_vendor_by_uid(self.context, vendor_uid)
            user = plone.api.user.get_current()
            if not user.checkPermission(permissions.ModifyOrders, vendor):
                return False
        else:
            vendor_uids = get_vendor_uids_for()
            if not vendor_uids:
                return False
        return True

    def render_salaried(self, colname, record):
        if not self.check_modify_order(record):
            salaried = OrderData(self.context, order=record).salaried
            return translate(vocabs.salaried_vocab()[salaried], context=self.request)
        return OrderSalariedDropdown(self.context, self.request, record).render()

    def render_state(self, colname, record):
        if not self.check_modify_order(record):
            state = OrderData(self.context, order=record).state
            return translate(vocabs.state_vocab()[state], context=self.request)
        return OrderStateDropdown(self.context, self.request, record).render()

    @property
    def ajaxurl(self):
        params = [
            ("vendor", self.request.form.get("vendor")),
            ("customer", self.request.form.get("customer")),
            ("state", self.request.form.get("state")),
            ("salaried", self.request.form.get("salaried")),
        ]
        query = six.moves.urllib.parse.urlencode(dict([it for it in params if it[1]]))
        query = query and u"?{0}".format(query) or ""
        return u"{0:s}/{1:s}{2:s}".format(
            self.context.absolute_url(), self.data_view_name, query
        )

    def __call__(self):
        # check if authenticated user is vendor
        if not get_vendors_for():
            raise Unauthorized
        # disable diazo theming if ajax call
        if "_" in self.request.form:
            self.request.response.setHeader("X-Theme-Disabled", "True")
        return super(OrdersTable, self).__call__()


class MyOrdersTable(OrdersTableBase):
    data_view_name = "@@myordersdata"

    def render_order_actions(self, colname, record):
        tag = Tag(Translate(self.request))
        base_url = self.context.absolute_url()

        # view order
        view_order_target = "%s?uid=%s" % (base_url, str(record.attrs["uid"]))
        view_order_attrs = {
            "ajax:bind": "click",
            "ajax:target": view_order_target,
            "ajax:overlay": "myorder",
            "href": "",
            "title": _("view_order", default=u"View Order"),
        }
        order_icon_url = "%s/++resource++bda.plone.orders/order.png" % base_url
        view_order_icon = tag("img", src=order_icon_url)
        view_order = tag("a", view_order_icon, **view_order_attrs)

        # view invoice
        view_invoice_target = "%s?uid=%s" % (base_url, str(record.attrs["uid"]))
        view_invoice_attrs = {
            "ajax:bind": "click",
            "ajax:target": view_invoice_target,
            "ajax:overlay": "myinvoice",
            "ajax:overlay-css": "invoice_overlay",
            "href": "",
            "title": _("view_invoice", default=u"View Invoice"),
        }
        invoice_icon_url = "%s/++resource++bda.plone.orders/invoice.png" % base_url
        view_invoice_icon = tag("img", src=invoice_icon_url)
        view_invoice = tag("a", view_invoice_icon, **view_invoice_attrs)

        # return joined actions
        return view_order + view_invoice

    def __call__(self):
        # disable diazo theming if ajax call
        if "_" in self.request.form:
            self.request.response.setHeader("X-Theme-Disabled", "True")
        return super(MyOrdersTable, self).__call__()


class OrdersData(OrdersTable, TableData):
    soup_name = "bda_plone_orders_orders"
    search_text_index = "text"

    def _get_buyables_in_context(self):
        catalog = plone.api.portal.get_tool("portal_catalog")
        path = "/".join(self.context.getPhysicalPath())
        brains = catalog(path=path, object_provides=IBuyable.__identifier__)
        for brain in brains:
            yield brain.UID

    def query(self, soup):
        # fetch user vendor uids
        vendor_uids = get_vendor_uids_for()
        # filter by given vendor uid or user vendor uids
        vendor_uid = self.request.form.get("vendor")
        if vendor_uid:
            vendor_uid = uuid.UUID(vendor_uid)
            # raise if given vendor uid not in user vendor uids
            if vendor_uid not in vendor_uids:
                raise Unauthorized
            query = Any("vendor_uids", [vendor_uid])
        else:
            query = Any("vendor_uids", vendor_uids)

        # filter by customer if given
        customer = self.request.form.get("customer")
        if customer:
            query = query & Eq("creator", customer)

        # Filter by state if given
        state = self.request.form.get("state")
        if state:
            query = query & Eq("state", state)

        # Filter by salaried if given
        salaried = self.request.form.get("salaried")
        if salaried:
            query = query & Eq("salaried", salaried)

        # filter by search term if given
        term = self.request.form["sSearch"]
        if six.PY2:
            term = term.decode("utf-8")
        if term:
            # append * for proper fulltext search
            term += "*"
            query = query & Contains(self.search_text_index, term)
        # get buyable uids for given context, get all buyables on site root
        # use explicit IPloneSiteRoot to make it play nice with lineage
        if not IPloneSiteRoot.providedBy(self.context):
            buyable_uids = self._get_buyables_in_context()
            query = query & Any("buyable_uids", buyable_uids)
        # query orders and return result
        sort = self.sort()
        res = soup.lazy(
            query, sort_index=sort["index"], reverse=sort["reverse"], with_size=True
        )
        length = next(res)
        return length, res


class MyOrdersData(MyOrdersTable, TableData):
    soup_name = "bda_plone_orders_orders"
    search_text_index = "text"

    def query(self, soup):
        query = Eq("creator", plone.api.user.get_current().getId())
        # filter by search term if given
        term = self.request.form["sSearch"]
        if six.PY2:
            term = term.decode("utf-8")
        if term:
            # append * for proper fulltext search
            term += "*"
            query = query & Contains(self.search_text_index, term)
        # query orders and return result
        sort = self.sort()
        res = soup.lazy(
            query, sort_index=sort["index"], reverse=sort["reverse"], with_size=True
        )
        length = next(res)
        return length, res



# -*- coding: utf-8 -*-
from bda.plone.orders import interfaces as ifaces
from bda.plone.orders.interfaces import IDynamicMailTemplateLibrary
from bda.plone.orders.interfaces import IDynamicMailTemplateLibraryStorage
from BTrees.OOBTree import OOBTree
from zope.annotation import IAnnotations
from zope.component import queryAdapter
from zope.interface import implementer


###############################################################################
# text templates
###############################################################################

# EN ##########################################################################

ORDER_SUBJECT_EN = u"Order %s received."

ORDER_BODY_EN = u"""
Date: %(date)s

Thank you for your order:

Ordernumber: %(ordernumber)s
Order details: %(portal_url)s/@@showorder?ordernumber=%(ordernumber)s

Invoice: %(portal_url)s/@@showinvoice?ordernumber=%(ordernumber)s

Personal Data:
Name: %(personal_data.firstname)s %(personal_data.lastname)s
Company: %(personal_data.company)s
Phone: %(personal_data.phone)s
Email: %(personal_data.email)s

Address:
Street: %(billing_address.street)s
ZIP: %(billing_address.zip)s
City: %(billing_address.city)s
Country: %(billing_address.country)s
%(delivery_address)s
Comment:
%(order_comment.comment)s

Ordered items:
%(item_listing)s

%(order_summary)s%(global_text)s%(payment_text)s
"""

RESERVATION_SUBJECT_EN = u"Reservation %s received."

RESERVATION_BODY_EN = u"""
Date: %(date)s

Thank you for your reservation:

Ordernumber: %(ordernumber)s
Reservation details: %(portal_url)s/@@showorder?ordernumber=%(ordernumber)s

Invoice: %(portal_url)s/@@showinvoice?ordernumber=%(ordernumber)s

Personal Data:
Name: %(personal_data.firstname)s %(personal_data.lastname)s
Company: %(personal_data.company)s
Phone: %(personal_data.phone)s
Email: %(personal_data.email)s

Address:
Street: %(billing_address.street)s
ZIP: %(billing_address.zip)s
City: %(billing_address.city)s
Country: %(billing_address.country)s
%(delivery_address)s
Comment:
%(order_comment.comment)s

Ordered items:
%(item_listing)s

Reserved items:
%(reserved_item_listing)s

%(order_summary)s%(global_text)s%(payment_text)s
"""

DELIVERY_ADDRESS_EN = u"""
Delivery Address:
Name: %(delivery_address.firstname)s %(delivery_address.lastname)s
Company: %(delivery_address.company)s
Street: %(delivery_address.street)s
ZIP: %(delivery_address.zip)s
City: %(delivery_address.city)s
Country: %(delivery_address.country)s
"""

CANCELLED_BOOKING_SUBJECT_EN = u"Cancelled one booking of Order %s."

CANCELLED_BOOKING_BODY_EN = u"""
Date: %(date)s

One of your bookings was cancelled.

Ordernumber: %(ordernumber)s
Cancelled item: %(booking_cancelled_title)s

Order details: %(portal_url)s/@@showorder?ordernumber=%(ordernumber)s
"""

BOOKING_RESERVED_TO_ORDERED_SUBJECT_EN = u"Reservation of %s is now available."

BOOKING_RESERVED_TO_ORDERED_BODY_EN = u"""
Date: %(date)s

Your reserved item is now available and our order is being processed.

Ordernumber: %(ordernumber)s
Booked item: %(booking_reserved_to_ordered_title)s

Order details: %(portal_url)s/@@showorder?ordernumber=%(ordernumber)s
"""

STOCK_THRESHOLD_REACHED_SUBJECT_EN = (
    u"Order %s has products that are getting out of stock."  # noqa
)

STOCK_THRESHOLD_REACHED_BODY_EN = u"""
Date: %(date)s

Products getting out of stock:
%(stock_threshold_reached_text)s
"""

# DE ##########################################################################

ORDER_SUBJECT_DE = u"Bestellung %s erhalten."

RESERVATION_SUBJECT_DE = u"Reservierung %s erhalten."

ORDER_BODY_DE = u"""
Datum: %(date)s

Besten Dank für Ihre Bestellung:

Bestellnummer: %(ordernumber)s
Details zur Bestellung: %(portal_url)s/@@showorder?ordernumber=%(ordernumber)s

Rechnung: %(portal_url)s/@@showinvoice?ordernumber=%(ordernumber)s

Persönliche Angaben:
Name: %(personal_data.firstname)s %(personal_data.lastname)s
Firma: %(personal_data.company)s
Telefon: %(personal_data.phone)s
E-Mail: %(personal_data.email)s

Adresse:
Strasse: %(billing_address.street)s
Postleitzahl: %(billing_address.zip)s
Ort: %(billing_address.city)s
Land: %(billing_address.country)s
%(delivery_address)s
Kommentar:
%(order_comment.comment)s

Bestellte Artikel:
%(item_listing)s

%(order_summary)s%(global_text)s%(payment_text)s
"""

RESERVATION_BODY_DE = u"""
Datum: %(date)s

Besten Dank für Ihre Reservierung:

Bestellnummer: %(ordernumber)s
Details zur Reservierung: %(portal_url)s/@@showorder?ordernumber=%(ordernumber)s

Rechnung: %(portal_url)s/@@showinvoice?ordernumber=%(ordernumber)s

Persönliche Angaben:
Name: %(personal_data.firstname)s %(personal_data.lastname)s
Firma: %(personal_data.company)s
Telefon: %(personal_data.phone)s
E-Mail: %(personal_data.email)s

Adresse:
Strasse: %(billing_address.street)s
Postleitzahl: %(billing_address.zip)s
Ort: %(billing_address.city)s
Land: %(billing_address.country)s
%(delivery_address)s
Kommentar:
%(order_comment.comment)s

Bestellte Artikel:
%(item_listing)s

Reservierte Artikel:
%(reserved_item_listing)s

%(order_summary)s%(global_text)s%(payment_text)s
"""  # noqa

DELIVERY_ADDRESS_DE = u"""
Lieferadresse:
Name: %(delivery_address.firstname)s %(delivery_address.lastname)s
Firma: %(delivery_address.company)s
Strasse: %(delivery_address.street)s
Postleitzahl: %(delivery_address.zip)s
Ort: %(delivery_address.city)s
Land: %(delivery_address.country)s
"""

CANCELLED_BOOKING_SUBJECT_DE = u"Stornierung eines Artikels in Bestellung %s."

CANCELLED_BOOKING_BODY_DE = u"""
Datum: %(date)s

Eine Ihrer bestellten Artikel wurde storniert.

Bestellnummer: %(ordernumber)s
Stornierter Artikel: %(booking_cancelled_title)s

Details zur Bestellung: %(portal_url)s/@@showorder?ordernumber=%(ordernumber)s
"""

BOOKING_RESERVED_TO_ORDERED_SUBJECT_DE = u"Reservierung %s ist nun verfügbar."

BOOKING_RESERVED_TO_ORDERED_BODY_DE = u"""
Datum: %(date)s

Ihr reservierter Artikel ist verfügbar geworden und Ihre Bestellung wird nun bearbeitet.

Bestellnummer: %(ordernumber)s
Bestellter Artikel: %(booking_reserved_to_ordered_title)s

Details zur Bestellung: %(portal_url)s/@@showorder?ordernumber=%(ordernumber)s
"""  # noqa

# FR ##########################################################################

ORDER_SUBJECT_FR = u"votre commande %s."

RESERVATION_SUBJECT_FR = u"votre réservation %s."

ORDER_BODY_FR = u"""
Date: %(date)s

Nous vous remercions pour votre commande:

No. de commande: %(ordernumber)s
%(portal_url)s/@@showorder?ordernumber=%(ordernumber)s

Le calcul: %(portal_url)s/@@showinvoice?ordernumber=%(ordernumber)s

Données personnelles:
Nom: %(personal_data.firstname)s %(personal_data.lastname)s
Entreprise: %(personal_data.company)s
Téléphone: %(personal_data.phone)s
E-Mail: %(personal_data.email)s

Adresse:
Rue: %(billing_address.street)s
No. Postal: %(billing_address.zip)s
Localité: %(billing_address.city)s
Pays: %(billing_address.country)s
%(delivery_address)s
Commentaires:
%(order_comment.comment)s

Produit commandé:
%(item_listing)s

%(order_summary)s%(global_text)s%(payment_text)s
"""

RESERVATION_BODY_FR = u"""
Date: %(date)s

Nous vous remercions pour votre réservation:

No. de commande: %(ordernumber)s
%(portal_url)s/@@showorder?ordernumber=%(ordernumber)s

Le calcul: %(portal_url)s/@@showinvoice?ordernumber=%(ordernumber)s

Données personnelles:
Nom: %(personal_data.firstname)s %(personal_data.lastname)s
Entreprise: %(personal_data.company)s
Téléphone: %(personal_data.phone)s
E-Mail: %(personal_data.email)s

Adresse:
Rue: %(billing_address.street)s
No. Postal: %(billing_address.zip)s
Localité: %(billing_address.city)s
Pays: %(billing_address.country)s
%(delivery_address)s
Commentaires:
%(order_comment.comment)s

Produit commandé:
%(item_listing)s

Produit réservés:
%(reserved_item_listing)s

%(order_summary)s%(global_text)s%(payment_text)s
"""

DELIVERY_ADDRESS_FR = u"""
Adresse de livraison:
Nom: %(delivery_address.firstname)s %(delivery_address.lastname)s
Entreprise: %(delivery_address.company)s
Rue: %(delivery_address.street)s
No. Postal: %(delivery_address.zip)s
Localité: %(delivery_address.city)s
Pays: %(delivery_address.country)s
"""

CANCELLED_BOOKING_SUBJECT_FR = u"Cancelled one booking of Order %s."

CANCELLED_BOOKING_BODY_FR = u"""
Date: %(date)s

One of your bookings was cancelled.

Ordernumber: %(ordernumber)s
Cancelled item: %(booking_cancelled_title)s

Order details: %(portal_url)s/@@showorder?ordernumber=%(ordernumber)s
"""

# IT ##########################################################################

ORDER_SUBJECT_IT = u"Il tuo ordine %s."

RESERVATION_SUBJECT_IT = u"Il tuo prenotazione %s."

ORDER_BODY_IT = u"""
Data: %(date)s

Grazie per l'ordine effettuato:

Numero d'ordine: %(ordernumber)s
%(portal_url)s/@@showorder?ordernumber=%(ordernumber)s

La fattura: %(portal_url)s/@@showinvoice?ordernumber=%(ordernumber)s

Dati personali:
Nome: %(personal_data.firstname)s %(personal_data.lastname)s
Ditta: %(personal_data.company)s
Telefono: %(personal_data.phone)s

Indirizzo:
Via: %(billing_address.street)s
CAP: %(billing_address.zip)s
Città: %(billing_address.city)s
Nazione: %(billing_address.country)s
%(delivery_address)s
Commento:
%(order_comment.comment)s

Articolo ordinato:
%(item_listing)s

%(order_summary)s%(global_text)s%(payment_text)s
"""

RESERVATION_BODY_IT = u"""
Data: %(date)s

Grazie per l'prenotazione effettuato:

Numero d'ordine: %(ordernumber)s
%(portal_url)s/@@showorder?ordernumber=%(ordernumber)s

La fattura: %(portal_url)s/@@showinvoice?ordernumber=%(ordernumber)s

Dati personali:
Nome: %(personal_data.firstname)s %(personal_data.lastname)s
Ditta: %(personal_data.company)s
Telefono: %(personal_data.phone)s

Indirizzo:
Via: %(billing_address.street)s
CAP: %(billing_address.zip)s
Città: %(billing_address.city)s
Nazione: %(billing_address.country)s
%(delivery_address)s
Commento:
%(order_comment.comment)s

Articolo ordinato:
%(item_listing)s

Articolo riservato:
%(reserved_item_listing)s

%(order_summary)s%(global_text)s%(payment_text)s
"""

DELIVERY_ADDRESS_IT = u"""
Indirizzo di spedizione:
Nome: %(delivery_address.firstname)s %(delivery_address.lastname)s
Ditta: %(delivery_address.company)s
Via: %(delivery_address.street)s
CAP: %(delivery_address.zip)s
Città: %(delivery_address.city)s
Nazione: %(delivery_address.country)s
"""

CANCELLED_BOOKING_SUBJECT_IT = u"Cancelled one booking of Order %s."

CANCELLED_BOOKING_BODY_IT = u"""
Date: %(date)s

One of your bookings was cancelled.

Ordernumber: %(ordernumber)s
Cancelled item: %(booking_cancelled_title)s

Order details: %(portal_url)s/@@showorder?ordernumber=%(ordernumber)s
"""

# NO ##########################################################################

ORDER_SUBJECT_NO = u"Bestilling %s mottatt."

RESERVATION_SUBJECT_NO = u"Bestilling %s mottatt."

ORDER_BODY_NO = u"""
Dato: %(date)s

Takk for din bestilling:

Ordernummer: %(ordernumber)s
%(portal_url)s/@@showorder?ordernumber=%(ordernumber)s

Regning: %(portal_url)s/@@showinvoice?ordernumber=%(ordernumber)s

Personlig info:
Navn: %(personal_data.firstname)s %(personal_data.lastname)s
Firma: %(personal_data.company)s
Telefon: %(personal_data.phone)s
Epost: %(personal_data.email)s

Adr:
Gate/Vei: %(billing_address.street)s
Postnr.: %(billing_address.zip)s
Poststed: %(billing_address.city)s
Land: %(billing_address.country)s
%(delivery_address)s
Kommentar:
%(order_comment.comment)s

Bestilte produkter:
%(item_listing)s

%(order_summary)s%(global_text)s%(payment_text)s
"""

RESERVATION_BODY_NO = u"""
Dato: %(date)s

Takk for din bestilling:

Ordernummer: %(ordernumber)s
%(portal_url)s/@@showorder?ordernumber=%(ordernumber)s

Regning: %(portal_url)s/@@showinvoice?ordernumber=%(ordernumber)s

Personlig info:
Navn: %(personal_data.firstname)s %(personal_data.lastname)s
Firma: %(personal_data.company)s
Telefon: %(personal_data.phone)s
Epost: %(personal_data.email)s

Adr:
Gate/Vei: %(billing_address.street)s
Postnr.: %(billing_address.zip)s
Poststed: %(billing_address.city)s
Land: %(billing_address.country)s
%(delivery_address)s
Kommentar:
%(order_comment.comment)s

Bestilte produkter:
%(item_listing)s

Reserverte produkter:
%(reserved_item_listing)s

%(order_summary)s%(global_text)s%(payment_text)s
"""

DELIVERY_ADDRESS_NO = u"""
Leveringsadr.:
Navn: %(delivery_address.firstname)s %(delivery_address.lastname)s
Firma: %(delivery_address.company)s
gate/Vei: %(delivery_address.street)s
Postnr.: %(delivery_address.zip)s
Poststed: %(delivery_address.city)s
Land: %(delivery_address.country)s
"""

CANCELLED_BOOKING_SUBJECT_NO = u"Cancelled one booking of Order %s."

CANCELLED_BOOKING_BODY_NO = u"""
Date: %(date)s

One of your bookings was cancelled.

Ordernumber: %(ordernumber)s
Cancelled item: %(booking_cancelled_title)s

Order details: %(portal_url)s/@@showorder?ordernumber=%(ordernumber)s
"""


###############################################################################
# text template registry
###############################################################################

ORDER_TEMPLATES = {
    "en": {
        "subject": ORDER_SUBJECT_EN,
        "body": ORDER_BODY_EN,
        "delivery_address": DELIVERY_ADDRESS_EN,
    },
    "de": {
        "subject": ORDER_SUBJECT_DE,
        "body": ORDER_BODY_DE,
        "delivery_address": DELIVERY_ADDRESS_DE,
    },
    "de-ch": {
        "subject": ORDER_SUBJECT_DE,
        "body": ORDER_BODY_DE,
        "delivery_address": DELIVERY_ADDRESS_DE,
    },
    "fr": {
        "subject": ORDER_SUBJECT_FR,
        "body": ORDER_BODY_FR,
        "delivery_address": DELIVERY_ADDRESS_FR,
    },
    "it": {
        "subject": ORDER_SUBJECT_IT,
        "body": ORDER_BODY_IT,
        "delivery_address": DELIVERY_ADDRESS_IT,
    },
    "no": {
        "subject": ORDER_SUBJECT_NO,
        "body": ORDER_BODY_NO,
        "delivery_address": DELIVERY_ADDRESS_NO,
    },
}

RESERVATION_TEMPLATES = {
    "en": {
        "subject": RESERVATION_SUBJECT_EN,
        "body": RESERVATION_BODY_EN,
        "delivery_address": DELIVERY_ADDRESS_EN,
    },
    "de": {
        "subject": RESERVATION_SUBJECT_DE,
        "body": RESERVATION_BODY_DE,
        "delivery_address": DELIVERY_ADDRESS_DE,
    },
    "de-ch": {
        "subject": RESERVATION_SUBJECT_DE,
        "body": RESERVATION_BODY_DE,
        "delivery_address": DELIVERY_ADDRESS_DE,
    },
    "fr": {
        "subject": RESERVATION_SUBJECT_FR,
        "body": RESERVATION_BODY_FR,
        "delivery_address": DELIVERY_ADDRESS_FR,
    },
    "it": {
        "subject": RESERVATION_SUBJECT_IT,
        "body": RESERVATION_BODY_IT,
        "delivery_address": DELIVERY_ADDRESS_IT,
    },
    "no": {
        "subject": RESERVATION_SUBJECT_NO,
        "body": RESERVATION_BODY_NO,
        "delivery_address": DELIVERY_ADDRESS_NO,
    },
}

CANCELLED_BOOKING_TEMPLATES = {
    "en": {"subject": CANCELLED_BOOKING_SUBJECT_EN, "body": CANCELLED_BOOKING_BODY_EN},
    "de": {"subject": CANCELLED_BOOKING_SUBJECT_DE, "body": CANCELLED_BOOKING_BODY_DE},
    "de-ch": {
        "subject": CANCELLED_BOOKING_SUBJECT_DE,
        "body": CANCELLED_BOOKING_BODY_DE,
    },
    "fr": {"subject": CANCELLED_BOOKING_SUBJECT_FR, "body": CANCELLED_BOOKING_BODY_FR},
    "it": {"subject": CANCELLED_BOOKING_SUBJECT_IT, "body": CANCELLED_BOOKING_BODY_IT},
    "no": {"subject": CANCELLED_BOOKING_SUBJECT_NO, "body": CANCELLED_BOOKING_BODY_NO},
}

BOOKING_RESERVED_TO_ORDERED_TEMPLATES = {
    "en": {
        "subject": BOOKING_RESERVED_TO_ORDERED_SUBJECT_EN,
        "body": BOOKING_RESERVED_TO_ORDERED_BODY_EN,
    },
    "de": {
        "subject": BOOKING_RESERVED_TO_ORDERED_SUBJECT_DE,
        "body": BOOKING_RESERVED_TO_ORDERED_BODY_DE,
    },
    "de-ch": {
        "subject": BOOKING_RESERVED_TO_ORDERED_SUBJECT_DE,
        "body": BOOKING_RESERVED_TO_ORDERED_BODY_DE,
    },
}

STOCK_THRESHOLD_REACHED_TEMPLATES = {
    "en": {
        "subject": STOCK_THRESHOLD_REACHED_SUBJECT_EN,
        "body": STOCK_THRESHOLD_REACHED_BODY_EN,
    }
}


def _get_templates(context, TPL, default="en"):
    lang = context.restrictedTraverse("@@plone_portal_state").language()
    return TPL.get(lang, TPL[default])


def get_order_templates(context):
    return _get_templates(context, ORDER_TEMPLATES)


def get_reservation_templates(context):
    return _get_templates(context, RESERVATION_TEMPLATES)


def get_booking_cancelled_templates(context):
    return _get_templates(context, CANCELLED_BOOKING_TEMPLATES)


def get_booking_reserved_to_ordered_templates(context):
    return _get_templates(context, BOOKING_RESERVED_TO_ORDERED_TEMPLATES)


def get_stock_threshold_reached_templates(context):
    return _get_templates(context, STOCK_THRESHOLD_REACHED_TEMPLATES)


###############################################################################
# dynamic mail templates
###############################################################################

# list of template attributes which are required. by default, no attributes are
# required.
REQUIRED_TEMPLATE_ATTRS = list()


# dictionary with attributes valid in mail template as keys, values are used
# for template validation
DEFAULT_TEMPLATE_ATTRS = {
    "created": "14.2.2014 14:42",
    "ordernumber": "123456",
    "salaried": ifaces.SALARIED_NO,
    "state": ifaces.STATE_NEW,
    "personal_data.company": "ACME LTD.",
    "personal_data.email": "max.mustermann@example.com",
    "personal_data.gender": "male",
    "personal_data.firstname": "Max",
    "personal_data.phone": "+43 123 456 78 90",
    "personal_data.lastname": "Mustermann",
    "billing_address.city": "Springfield",
    "billing_address.country": "Austria",
    "billing_address.street": "Musterstrasse",
    "billing_address.zip": "1234",
    "order_comment.comment": "Comment",
    "payment_selection.payment": "six_payment",
}


class DynamicMailTemplate(object):
    """Dynamic Mail Template based on str.format
    """

    def __init__(self, required=[], defaults={}):
        """Initialize a new template

        required
            a list of keys which are required in the data to be rendered.
            all other values are taken from defaults if not provided.

        defaults
            a complete set of values for the template. used for validation,
            so it has to include all required as well.
        """
        for key in required:
            if key in defaults:
                continue
            raise ValueError(
                u"All required must be in defaults too, missing: " u"{0}".format(key)
            )
        self.required = required
        self.defaults = defaults

    def normalized(self, keys=[], indict={}):
        if keys and indict:
            raise ValueError("Only one kwargs please.")
        if keys:
            result = []
            for key in keys:
                result.append(key.replace(".", "_"))
            return result
        if indict:
            result = {}
            for key, value in indict.items():
                result[key.replace(".", "_")] = value
            return result
        raise ValueError("Only one kwargs please.")

    def validate(self, template):
        """validates if the template can be rendered.

        uses default values to achieve this

        template
            a unicode string meant to be rendered using python string format
            method
        """
        assert isinstance(template, six.text_type), "template must be unicode"
        try:
            self(template, self.defaults)
        except KeyError as e:
            return False, u'Variable "{0}" is not available.'.format(e.message)
        except Exception as exc:
            return False, str(exc)
        return True, ""

    def __call__(self, template, data):
        """render template with data
        """
        assert isinstance(template, six.text_type), "template must be unicode"
        for key in self.required:
            if key not in data:
                raise KeyError(u"Required key {0} is missing.".format(key))
        return template.format(**self.normalized(indict=data))


DYNAMIC_MAIL_LIBRARY_KEY = "bda.plone.order.dynamic_mail_lib"


@implementer(IDynamicMailTemplateLibrary)
class DynamicMailTemplateLibraryAquierer(object):
    def __init__(self, context):
        self.context = context

    def _parent(self):
        if not hasattr(self.context, "__parent__"):
            return None
        if self.context.__parent__:
            dmt_lib = queryAdapter(self.context.__parent__, IDynamicMailTemplateLibrary)
            return dmt_lib

    def keys(self):
        parent = self._parent()
        if parent is None:
            return []
        return list(parent.keys())

    def __getitem__(self, name):
        parent = self._parent()
        if parent is not None:
            return parent[name]
        raise KeyError("Can not aquire key %s" % name)

    def __setitem__(self, name, template):
        raise NotImplementedError("acquierer do not set on parent (permissions)")

    def __delitem__(self, name):
        raise NotImplementedError("acquierer do not delete on parent (permissions)")


@implementer(IDynamicMailTemplateLibraryStorage)
class DynamicMailTemplateLibraryStorage(DynamicMailTemplateLibraryAquierer):
    @property
    def _storage(self):
        annotations = IAnnotations(self.context)
        if DYNAMIC_MAIL_LIBRARY_KEY not in annotations:
            annotations[DYNAMIC_MAIL_LIBRARY_KEY] = OOBTree()
        return annotations[DYNAMIC_MAIL_LIBRARY_KEY]

    def direct_keys(self):
        return [_ for _ in self._storage.keys()]

    def keys(self):
        result = self.direct_keys()
        parent_keys = list(super(DynamicMailTemplateLibraryStorage, self).keys())
        for key in parent_keys:
            if key not in result:  # child wins
                result.append(key)
        return result

    def __getitem__(self, name):
        try:
            return self._storage[name]
        except KeyError:
            return super(DynamicMailTemplateLibraryStorage, self).__getitem__(name)

    def __setitem__(self, name, template):
        self._storage[name] = template

    def __delitem__(self, name):
        del self._storage[name]



# -*- coding: utf-8 -*-
from bda.plone.cart.restapi.service import TraversingService
from bda.plone.orders.common import get_vendor_uids_for
from bda.plone.orders.datamanagers.booking import BookingData
from bda.plone.orders.interfaces import workflow
from plone.protect.interfaces import IDisableCSRFProtection
from plone.restapi.interfaces import ISerializeToJson
from zope.component import getMultiAdapter
from zope.interface import alsoProvides


class BookingService(TraversingService):
    """Single Booking"""

    def process_request(self, booking_data):
        pass

    def reply(self):
        if len(self.params) != 1:
            raise Exception(
                "Must supply exactly one UID (unique identifier of"
                "the booking) to be retrieved as last part of the URL ."
            )
        booking_data = BookingData(self.context, uid=self.params[0])
        if booking_data.booking is None:
            raise ValueError("Can not find given UID '{0}'".format(self.params[0]))
        self.process_request(booking_data)
        serializer = getMultiAdapter((booking_data, self.request), ISerializeToJson)
        return serializer()


class BookingUpdateService(BookingService):
    """Single Booking update: state and salaried"""

    def process_request(self, booking_data):
        # Disable CSRF protection
        alsoProvides(self.request, IDisableCSRFProtection)
        salaried = self.request.form.get("salaried", booking_data.salaried)
        if salaried not in [
            workflow.SALARIED_YES,
            workflow.SALARIED_NO,
            workflow.SALARIED_FAILED,
        ]:
            raise Exception(
                "salaried must be one out of: {0}, {1}, {2}.".format(
                    workflow.SALARIED_YES,
                    workflow.SALARIED_NO,
                    workflow.SALARIED_FAILED,
                )
            )
        if salaried != booking_data.salaried:
            booking_data.salaried = salaried

        state = self.request.form.get("state", booking_data.state)
        if state not in [
            workflow.STATE_CANCELLED,
            workflow.STATE_FINISHED,
            workflow.STATE_NEW,
            workflow.STATE_PROCESSING,
            workflow.STATE_RESERVED,
        ]:
            raise Exception(
                "state must be one out of: {0}, {1}, {2}, {3}, {4}.".format(
                    workflow.STATE_CANCELLED,
                    workflow.STATE_FINISHED,
                    workflow.STATE_NEW,
                    workflow.STATE_PROCESSING,
                    workflow.STATE_RESERVED,
                )
            )
        if state != booking_data.state:
            booking_data.state = state



# -*- coding: utf-8 -*-
from bda.plone.cart.utils import ascur
from bda.plone.cart.utils import get_catalog_brain
from bda.plone.checkout.interfaces import ICheckoutEvent
from bda.plone.checkout.interfaces import ICheckoutSettings
from bda.plone.orders import get_country_name
from bda.plone.orders import interfaces as ifaces
from bda.plone.orders import message_factory as _
from bda.plone.orders import vocabularies as vocabs
from bda.plone.orders.common import DT_FORMAT
from bda.plone.orders.datamanagers.order import OrderData
from bda.plone.orders.interfaces import IGlobalNotificationText
from bda.plone.orders.interfaces import IItemNotificationText
from bda.plone.orders.interfaces import INotificationSettings
from bda.plone.orders.interfaces import IPaymentText
from bda.plone.orders.mailtemplates import get_booking_cancelled_templates
from bda.plone.orders.mailtemplates import get_booking_reserved_to_ordered_templates
from bda.plone.orders.mailtemplates import get_order_templates
from bda.plone.orders.mailtemplates import get_reservation_templates
from bda.plone.orders.mailtemplates import get_stock_threshold_reached_templates
from chameleon import PageTemplateLoader
from decimal import Decimal
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formataddr
from plone import api
from Products.CMFPlone.utils import safe_unicode
from zope.component.hooks import getSite
from zope.globalrequest import getRequest
from zope.i18n import translate

import logging
import os
import textwrap


logger = logging.getLogger("bda.plone.orders")


NOTIFICATIONS = {}


POSSIBLE_TEMPLATE_CALLBACKS = [
    "booking_cancelled_title",
    "booking_reserved_to_ordered_title",
    "global_text",
    "item_listing",
    "reserved_item_listing",
    "order_summary",
    "payment_text",
    "stock_threshold_reached_text",
    "delivery_address",
]


def get_order_uid(event):
    uid = None
    if ICheckoutEvent.providedBy(event):
        uid = event.uid
    else:
        uid = event.order_uid
    return uid


def _indent(text, ind=5, width=80):
    """Helper indents text.
    """
    wrapped = textwrap.fill(safe_unicode(text), width, initial_indent=ind * u" ")
    return wrapped


def _process_template_cb(name, tpls, args, context, order_data):
    """Process template callback.

    The result of the template callback gets written to args.
    """
    cb_name = u"{0:s}_cb".format(name)
    if cb_name in tpls:
        args[name] = tpls[cb_name](context, order_data)


###############################################################################
# mail notification related data exraction
###############################################################################


def general_order_data(context, order_data):
    lang = context.restrictedTraverse("@@plone_portal_state").language()
    attrs = order_data.order.attrs
    data = dict(
        (safe_unicode(key), safe_unicode(value)) for (key, value) in list(attrs.items())
    )
    data["portal_url"] = getSite().absolute_url()
    data["date"] = attrs["created"].strftime(DT_FORMAT)
    data["salutation"] = translate(
        attrs.get("personal_data.gender", "dear"),
        domain="bda.plone.checkout",
        target_language=lang,
    )
    if data.get("billing_address.country", None):
        data["billing_address.country"] = get_country_name(
            data["billing_address.country"], lang=lang
        )
    if data.get("delivery_address.country", None):
        data["delivery_address.country"] = get_country_name(
            data["delivery_address.country"], lang=lang
        )
    return data


def order_item_data(context, booking):
    """Extract data for one listing item.
    """
    data = dict()
    data["title"] = safe_unicode(booking.attrs["title"])
    data["item_number"] = safe_unicode(booking.attrs["item_number"])
    data["comment"] = safe_unicode(booking.attrs["buyable_comment"])
    data["currency"] = safe_unicode(booking.attrs["currency"])
    data["buyable_count"] = Decimal(booking.attrs["buyable_count"])
    data["quantity_unit"] = booking.attrs["quantity_unit"]
    data["net"] = Decimal(booking.attrs["net"])
    data["net_total"] = Decimal(booking.attrs["net"]) * Decimal(booking.attrs["buyable_count"])
    data["vat"] = booking.attrs["vat"]
    data["gross"] = Decimal(booking.attrs["net"]) * (1 + booking.attrs["vat"] / 100)
    data["gross_total"] = data["gross"] * Decimal(booking.attrs["buyable_count"])
    data["discount_net"] = Decimal(booking.attrs["discount_net"])
    data["discount_gross"] = Decimal(booking.attrs["discount_net"]) * (1 + booking.attrs["vat"] / 100)
    data["state"] = state = safe_unicode(booking.attrs.get("state"))
    brain = get_catalog_brain(context, booking.attrs["buyable_uid"])
    buyable = brain.getObject()
    notificationtext = IItemNotificationText(buyable)
    text = None
    if state == ifaces.STATE_RESERVED:
        text = notificationtext.overbook_text
    elif state == ifaces.STATE_NEW:
        text = notificationtext.order_text
    data["notification"] = text
    return data


default_include_booking_states = (
    ifaces.STATE_FINISHED,
    ifaces.STATE_NEW,
    ifaces.STATE_PROCESSING,
)


def order_items_data(
    context, order_data, include_booking_states=default_include_booking_states
):
    """Extract item listing data.
    """
    data = list()
    for booking in order_data.bookings:
        if booking.attrs.get("state") not in include_booking_states:
            continue
        data.append(order_item_data(context, booking))
    return data


def order_summary_data(order_data):
    """Extract data for order summary.
    """
    data = dict()
    data["currency"] = order_data.currency
    cart_net = order_data.net
    data["cart_net"] = cart_net
    cart_vat = order_data.vat
    data["cart_vat"] = cart_vat
    discount_net = order_data.discount_net
    data["discount_net"] = discount_net
    discount_vat = order_data.discount_vat
    data["discount_vat"] = discount_vat
    discount_total = discount_net + discount_vat
    data["discount_total"] = discount_total
    attrs = order_data.order.attrs
    data["shipping_label"] = attrs["shipping_label"]
    data["shipping_description"] = attrs["shipping_description"]
    shipping_net = order_data.shipping_net
    data["shipping_net"] = shipping_net
    shipping_vat = order_data.shipping_vat
    data["shipping_vat"] = shipping_vat
    shipping_total = shipping_net + shipping_vat
    data["shipping_total"] = shipping_total
    cart_total = order_data.total
    data["cart_total"] = cart_total
    return data


def order_payment_data(order_data):
    data = dict()
    data["payment_method"] = payment = order_data.order.attrs["payment_method"]
    data["payment_text"] = IPaymentText(getSite()).payment_text(payment).strip()
    return data


def order_notifications(context, order_data):
    order_state = order_data.state
    notifications = set()
    for booking in order_data.bookings:
        brain = get_catalog_brain(context, booking.attrs["buyable_uid"])
        buyable = brain.getObject()
        notificationtext = IGlobalNotificationText(buyable)
        if order_state in (ifaces.STATE_RESERVED, ifaces.STATE_MIXED):
            # XXX: might need custom text for MIXED state
            text = notificationtext.global_overbook_text.strip()
            if text:
                notifications.add(text)
        elif order_state == ifaces.STATE_NEW:
            text = notificationtext.global_order_text.strip()
            if text:
                notifications.add(text)
    return list(notifications)


def stock_threshold_reached_items_data(items):
    ret = list()
    for item in items:
        ret.append(
            {
                "title": safe_unicode(item["title"]),
                "remaining_stock_available": item["remaining_stock_available"],
            }
        )
    return ret


###############################################################################
# mail notification
###############################################################################


class MailNotify(object):
    """Mail notifyer.
    """

    def __init__(self, context):
        self.context = context
        self.settings = INotificationSettings(self.context)

    def create_multipart_message(self, subject, from_, to, text, html):
        message = MIMEMultipart("alternative")
        message["Subject"] = subject
        message["From"] = from_
        message["To"] = to
        # According to RFC 2046, the last part of a multipart message, in this case
        # the HTML message, is best and preferred.
        message.attach(MIMEText(text, "plain", "utf-8"))
        message.attach(MIMEText(html, "html", "utf-8"))
        return message

    def send(self, subject, receiver, text, html=None):
        shop_manager_address = self.settings.admin_email
        if not shop_manager_address:
            raise ValueError("Shop manager address is missing in settings.")
        shop_manager_name = self.settings.admin_name
        if shop_manager_name:
            from_name = shop_manager_name
            mailfrom = formataddr((from_name, shop_manager_address))
        else:
            mailfrom = shop_manager_address
        if html is None:
            message = text
        else:
            message = self.create_multipart_message(
                subject, mailfrom, receiver, text, html
            )
        api.portal.send_email(
            recipient=receiver, sender=mailfrom, subject=subject, body=message
        )


def create_text_mail_body(context, order_data, templates):
    """Creates a rendered mail body

    context
        Some object in Plone which can be used as a context to acquire from.

    order_data
        Order-data instance.

    templates
        Dict with a bunch of cbs and the body template itself.
    """
    arguments = general_order_data(context, order_data)
    for name in POSSIBLE_TEMPLATE_CALLBACKS:
        _process_template_cb(name, templates, arguments, context, order_data)
    return templates["body"] % arguments


class ZPTTranslator(object):
    def __init__(self):
        self.request = getRequest()

    def __call__(
        self,
        msgid,
        domain=None,
        mapping=None,
        default=None,
        context=None,
        target_language=None,
    ):
        return translate(
            msgid, domain=domain, mapping=mapping, default=default, context=self.request
        )


MAIL_TEMPLATES_DIRECTORY = os.path.join(os.path.dirname(__file__), "mailtemplates")


def create_html_mail_body(context, template_name, template_data):
    """Creates a rendered mail body

    context
        Some object in Plone which can be used as a context to acquire from.

    template_name
        HTML template name as string.

    template_data
        Dict containing data passed to template.
    """
    templates = PageTemplateLoader(
        MAIL_TEMPLATES_DIRECTORY, translate=ZPTTranslator(), debug=True
    )
    template = templates["{}.pt".format(template_name)]
    template_data["ascur"] = ascur
    return template(**template_data)


def do_notify(context, order_data, receiver, templates, template_name, template_data):
    """Do mail notification.
    """
    attrs = order_data.order.attrs
    subject = templates["subject"] % attrs["ordernumber"]
    text = create_text_mail_body(context, order_data, templates)
    html = create_html_mail_body(context, template_name, template_data)
    mail_notify = MailNotify(context)
    try:
        mail_notify.send(subject, receiver, text, html=html)
    except Exception:
        msg = translate(
            _(
                "email_sending_failed",
                default=u"Failed to send notification to ${receiver}",
                mapping={"receiver": receiver},
            )
        )
        api.portal.show_message(message=msg, request=context.REQUEST)
        logger.exception("Email could not be sent.")


def do_notify_customer(context, order_data, templates, template_name, template_data):
    customer_address = order_data.order.attrs["personal_data.email"]
    do_notify(
        context, order_data, customer_address, templates, template_name, template_data
    )


def do_notify_shopmanager(context, order_data, templates, template_name, template_data):
    shop_manager_address = INotificationSettings(context).admin_email
    do_notify(
        context,
        order_data,
        shop_manager_address,
        templates,
        template_name,
        template_data,
    )


###############################################################################
# order success
###############################################################################

# template callbacks and helpers ##############################################


def create_order_listing_item(item_data):
    """Create text for one item in order.
    """
    # count
    text = u"{:4f}".format(item_data["buyable_count"])
    # title
    title = item_data["title"]
    comment = item_data["comment"]
    if comment:
        title = u"{} ({})".format(title, comment)
    text = u"{} {}".format(text, title)
    # item number
    item_number = item_data["item_number"]
    if item_number:
        text = u"{} ({})".format(text, item_number)
    # state
    state_text = u""
    if item_data["state"] == ifaces.STATE_RESERVED:
        state_text = u"({})".format(vocabs.state_vocab()[item_data["state"]])
    if state_text:
        text = u"{} {}".format(text, state_text)
    # price
    currency = item_data["currency"]
    net = item_data["net"]
    discount_net = item_data["discount_net"]
    if discount_net:
        net = net - discount_net
    text = u"{} {} {:0.2f}".format(text, currency, net)
    # notification text
    notification = item_data["notification"]
    if notification:
        return u"\n".join([text, _indent(notification)])
    return text


def create_order_listing(
    context, order_data, include_booking_states=default_include_booking_states
):
    """Create item listing text for notification mail.
    """
    items = []
    items_data = order_items_data(
        context, order_data, include_booking_states=include_booking_states
    )
    for item_data in items_data:
        items.append(create_order_listing_item(item_data))
    return u"\n".join(items)


def create_reserved_order_listing(context, order_data):
    """Create item listing text for notification mail containing reserved items
    of this order.
    """
    return create_order_listing(
        context, order_data, include_booking_states=(ifaces.STATE_RESERVED,)
    )


def create_order_summary(context, order_data):
    """Create cart summary text for notification mail.
    """
    # no costs at all
    if not order_data.total:
        return u""
    lines = []
    request = getRequest()
    summary_data = order_summary_data(order_data)
    # cart net and vat
    if summary_data["cart_net"]:
        # cart net
        lines.append(
            translate(
                _(
                    "order_summary_cart_net",
                    default=u"Net: ${value} ${currency}",
                    mapping={
                        "value": ascur(summary_data["cart_net"]),
                        "currency": summary_data["currency"],
                    },
                ),
                context=request,
            )
        )
        # cart vat
        lines.append(
            translate(
                _(
                    "order_summary_cart_vat",
                    default=u"VAT: ${value} ${currency}",
                    mapping={
                        "value": ascur(summary_data["cart_vat"]),
                        "currency": summary_data["currency"],
                    },
                ),
                context=request,
            )
        )
    # cart discount
    if summary_data["discount_net"]:
        # discount net
        lines.append(
            translate(
                _(
                    "order_summary_discount_net",
                    default=u"Discount Net: ${value} ${currency}",
                    mapping={
                        "value": ascur(summary_data["discount_net"]),
                        "currency": summary_data["currency"],
                    },
                ),
                context=request,
            )
        )
        # discount vat
        lines.append(
            translate(
                _(
                    "order_summary_discount_vat",
                    default=u"Discount VAT: ${value} ${currency}",
                    mapping={
                        "value": ascur(summary_data["discount_vat"]),
                        "currency": summary_data["currency"],
                    },
                ),
                context=request,
            )
        )
        # discount total
        lines.append(
            translate(
                _(
                    "order_summary_discount_total",
                    default=u"Discount Total: ${value} ${currency}",
                    mapping={
                        "value": ascur(summary_data["discount_total"]),
                        "currency": summary_data["currency"],
                    },
                ),
                context=request,
            )
        )
    # shipping costs
    if summary_data["shipping_net"]:
        # shiping label
        lines.append(
            translate(
                _(
                    "order_summary_shipping_label",
                    default=u"Shipping: ${label}",
                    mapping={
                        "label": translate(
                            summary_data["shipping_label"], context=request
                        )
                    },
                ),
                context=request,
            )
        )
        # shiping description
        lines.append(translate(summary_data["shipping_description"], context=request))
        # shiping net
        lines.append(
            translate(
                _(
                    "order_summary_shipping_net",
                    default=u"Shipping Net: ${value} ${currency}",
                    mapping={
                        "value": ascur(summary_data["shipping_net"]),
                        "currency": summary_data["currency"],
                    },
                ),
                context=request,
            )
        )
        # shiping vat
        lines.append(
            translate(
                _(
                    "order_summary_shipping_vat",
                    default=u"Shipping VAT: ${value} ${currency}",
                    mapping={
                        "value": ascur(summary_data["shipping_vat"]),
                        "currency": summary_data["currency"],
                    },
                ),
                context=request,
            )
        )
        # shiping total
        lines.append(
            translate(
                _(
                    "order_summary_shipping_total",
                    default=u"Shipping Total: ${value} ${currency}",
                    mapping={
                        "value": ascur(summary_data["shipping_total"]),
                        "currency": summary_data["currency"],
                    },
                ),
                context=request,
            )
        )
    # cart total
    lines.append(
        translate(
            _(
                "order_summary_cart_total",
                default=u"Total: ${value} ${currency}",
                mapping={
                    "value": ascur(summary_data["cart_total"]),
                    "currency": summary_data["currency"],
                },
            ),
            context=request,
        )
    )
    summary_title = translate(
        _("order_summary_label", default=u"Summary:"), context=request
    )
    summary_text = u"\n".join(lines)
    return u"\n{summary_title}\n{summary_text}\n".format(
        summary_title=summary_title, summary_text=summary_text
    )


def create_global_text(context, order_data):
    notifications = order_notifications(context, order_data)
    if notifications:
        global_text = u"\n\n".join(notifications)
        return u"\n\n{global_text}\n".format(global_text=global_text)
    return u""


def create_payment_text(context, order_data):
    payment_text = order_payment_data(order_data)["payment_text"]
    if payment_text:
        return u"\n\n{payment_text}\n".format(payment_text=payment_text)  # noqa
    return u""


def create_delivery_address(context, order_data):
    if order_data.order.attrs.get("delivery_address.alternative_delivery", None):
        templates = get_order_templates(context)
        delivery_address_template = templates["delivery_address"]
        arguments = general_order_data(context, order_data)
        return delivery_address_template % arguments
    return u""


# event handling ##############################################################


def dispatch_notify_order_success(event):
    for func in NOTIFICATIONS["order_success"]:
        func(event)


# notification ##############################################################

NOTIFY_ORDER_SUCCESS_TEMPLATE = "order_success"


def notify_order_success(event, who=None):
    """Send notification mail after order succeeded.
    """
    if who not in ["customer", "shopmanager"]:
        raise ValueError('kw "who" mus be one out of ("customer", "shopmanager")')
    order_data = OrderData(event.context, uid=get_order_uid(event))
    templates = dict()
    state = order_data.state
    if state in (ifaces.STATE_RESERVED, ifaces.STATE_MIXED):
        templates.update(get_reservation_templates(event.context))
        templates["reserved_item_listing_cb"] = create_reserved_order_listing
    else:
        templates.update(get_order_templates(event.context))
    templates["item_listing_cb"] = create_order_listing
    templates["order_summary_cb"] = create_order_summary
    templates["global_text_cb"] = create_global_text
    templates["payment_text_cb"] = create_payment_text
    templates["delivery_address_cb"] = create_delivery_address
    template_name = NOTIFY_ORDER_SUCCESS_TEMPLATE
    template_data = dict()
    template_data["order"] = general_order_data(event.context, order_data)
    template_data["items"] = order_items_data(event.context, order_data)
    template_data["reserved_items"] = order_items_data(
        event.context, order_data, include_booking_states=(ifaces.STATE_RESERVED,)
    )
    template_data["summary"] = order_summary_data(order_data)
    template_data["payment"] = order_payment_data(order_data)
    template_data["notifications"] = order_notifications(event.context, order_data)
    if who == "customer":
        do_notify_customer(
            event.context, order_data, templates, template_name, template_data
        )
    else:
        do_notify_shopmanager(
            event.context, order_data, templates, template_name, template_data
        )


def notify_order_success_customer(event):
    notify_order_success(event, who="customer")


def notify_order_success_shopmanager(event):
    notify_order_success(event, who="shopmanager")


NOTIFICATIONS["order_success"] = []
NOTIFICATIONS["order_success"].append(notify_order_success_customer)
NOTIFICATIONS["order_success"].append(notify_order_success_shopmanager)


###############################################################################
# checkout success
###############################################################################


def dispatch_notify_checkout_success(event):
    for func in NOTIFICATIONS["checkout_success"]:
        func(event)


def notify_checkout_success_customer(event):
    """Send notification mail after checkout succeed.
    """
    # if skip payment, do notification
    checkout_settings = ICheckoutSettings(event.context)
    if checkout_settings.skip_payment(get_order_uid(event)):
        notify_order_success(event, who="customer")


def notify_checkout_success_shopmanager(event):
    """Send notification mail after checkout succeed.
    """
    # if skip payment, do notification
    checkout_settings = ICheckoutSettings(event.context)
    if checkout_settings.skip_payment(get_order_uid(event)):
        notify_order_success(event, who="shopmanager")


NOTIFICATIONS["checkout_success"] = []
NOTIFICATIONS["checkout_success"].append(notify_checkout_success_customer)
NOTIFICATIONS["checkout_success"].append(notify_checkout_success_shopmanager)


###############################################################################
# payment success
###############################################################################


def dispatch_notify_payment_success(event):
    for func in NOTIFICATIONS["payment_success"]:
        func(event)


def notify_payment_success_customer(event):
    """Send notification mail after payment succeed.
    """
    notify_order_success(event, who="customer")


def notify_payment_success_shopmanager(event):
    """Send notification mail after payment succeed.
    """
    notify_order_success(event, who="shopmanager")


NOTIFICATIONS["payment_success"] = []
NOTIFICATIONS["payment_success"].append(notify_payment_success_customer)
NOTIFICATIONS["payment_success"].append(notify_payment_success_shopmanager)


###############################################################################
# booking cancelled
###############################################################################

BOOKING_CANCELLED_TITLE_ATTRIBUTE = "title"


def dispatch_notify_booking_cancelled(event):
    for func in NOTIFICATIONS["booking_cancelled"]:
        func(event)


class BookingCancelledTitleCB(object):
    def __init__(self, event):
        self.event = event

    def __call__(self, *args):
        return safe_unicode(self.event.booking_attrs[BOOKING_CANCELLED_TITLE_ATTRIBUTE])


BOOKING_CANCELLED_TEMPLATE = "booking_cancelled"


def notify_booking_cancelled(event, who=None):
    """Send notification mail after booking was cancelled.
    """
    order_data = OrderData(event.context, uid=get_order_uid(event))
    templates = dict()
    templates.update(get_booking_cancelled_templates(event.context))
    templates["booking_cancelled_title_cb"] = BookingCancelledTitleCB(event)
    template_name = BOOKING_CANCELLED_TEMPLATE
    template_data = dict()
    template_data["order"] = general_order_data(event.context, order_data)
    template_data["booking"] = dict()
    template_data["booking"]["title"] = BookingCancelledTitleCB(event)()
    if who == "customer":
        do_notify_customer(
            event.context, order_data, templates, template_name, template_data
        )
    elif who == "shopmanager":
        do_notify_shopmanager(
            event.context, order_data, templates, template_name, template_data
        )
    else:
        raise ValueError('kw "who" mus be one out of ("customer", "shopmanager")')


def notify_booking_cancelled_customer(event):
    notify_booking_cancelled(event, who="customer")


def notify_booking_cancelled_shopmanager(event):
    notify_booking_cancelled(event, who="shopmanager")


NOTIFICATIONS["booking_cancelled"] = []
NOTIFICATIONS["booking_cancelled"].append(notify_booking_cancelled_customer)
NOTIFICATIONS["booking_cancelled"].append(notify_booking_cancelled_shopmanager)


###############################################################################
# booking reserved to ordered
###############################################################################


def dispatch_notify_booking_reserved_to_ordered(event):
    for func in NOTIFICATIONS["booking_reserved_to_ordered"]:
        func(event)


BookingReservedToOrderedTitleCB = BookingCancelledTitleCB
BOOKING_RESERVED_TO_ORDERED_TEMPLATE = "booking_reserved_to_ordered"


def notify_booking_reserved_to_ordered(event, who=None):
    """Send notification mail after booking was changed from reserved to ordered.
    """
    order_data = OrderData(event.context, uid=get_order_uid(event))
    templates = dict()
    templates.update(get_booking_reserved_to_ordered_templates(event.context))
    templates["booking_reserved_to_ordered_title_cb"] = BookingReservedToOrderedTitleCB(
        event
    )  # noqa
    template_name = BOOKING_RESERVED_TO_ORDERED_TEMPLATE
    template_data = dict()
    template_data["order"] = general_order_data(event.context, order_data)
    template_data["booking"] = dict()
    template_data["booking"]["title"] = BookingReservedToOrderedTitleCB(event)()
    if who == "customer":
        do_notify_customer(
            event.context, order_data, templates, template_name, template_data
        )
    elif who == "shopmanager":
        do_notify_shopmanager(
            event.context, order_data, templates, template_name, template_data
        )
    else:
        raise ValueError('kw "who" mus be one out of ("customer", "shopmanager")')


def notify_booking_reserved_to_ordered_customer(event):
    notify_booking_reserved_to_ordered(event, who="customer")


def notify_booking_reserved_to_ordered_shopmanager(event):
    notify_booking_reserved_to_ordered(event, who="shopmanager")


NOTIFICATIONS["booking_reserved_to_ordered"] = []
NOTIFICATIONS["booking_reserved_to_ordered"].append(
    notify_booking_reserved_to_ordered_customer
)  # noqa
NOTIFICATIONS["booking_reserved_to_ordered"].append(
    notify_booking_reserved_to_ordered_shopmanager
)  # noqa


###############################################################################
# stock threshold reached
###############################################################################


def dispatch_notify_stock_threshold_reached(event):
    for func in NOTIFICATIONS["stock_threshold_reached"]:
        func(event)


class StockThresholdReachedCB(object):
    def __init__(self, event):
        self.event = event

    def __call__(self, *args):
        text = ""
        items = stock_threshold_reached_items_data(
            self.event.stock_threshold_reached_items
        )
        for item in items:
            title = safe_unicode(item["title"])
            remaining = item["remaining_stock_available"]
            text += u"{0} (Remaining stock: {1})\n".format(title, remaining)
        return text


STOCK_THRESHOLD_REACHED_TEMPLATE = "stock_threshold_reached"


def notify_stock_threshold_reached(event):
    """Send notification mail when item is getting out of stock.
    """
    order_data = OrderData(event.context, uid=get_order_uid(event))
    templates = dict()
    templates.update(get_stock_threshold_reached_templates(event.context))
    templates["stock_threshold_reached_text_cb"] = StockThresholdReachedCB(event)
    template_name = STOCK_THRESHOLD_REACHED_TEMPLATE
    template_data = dict()
    template_data["order"] = general_order_data(event.context, order_data)
    template_data["items"] = stock_threshold_reached_items_data(
        event.stock_threshold_reached_items
    )
    do_notify_shopmanager(
        event.context, order_data, templates, template_name, template_data
    )


NOTIFICATIONS["stock_threshold_reached"] = []
NOTIFICATIONS["stock_threshold_reached"].append(notify_stock_threshold_reached)



# -*- coding: utf-8 -*-
from bda.plone.cart.restapi.service import Service
from bda.plone.cart.restapi.service import TraversingService
from bda.plone.orders.common import get_vendor_uids_for
from bda.plone.orders.datamanagers.order import OrderData
from bda.plone.orders.datamanagers.orders import OrdersManager
from plone.restapi.interfaces import ISerializeToJson
from repoze.catalog.query import Any
from repoze.catalog.query import Eq
from zope.component import getMultiAdapter


class OrderService(TraversingService):
    """Single Order with bookings"""

    def reply(self):
        if len(self.params) != 1:
            raise Exception(
                "Must supply exactly one UID (unique identifier of"
                "the order) to be retrieved as last part of the URL ."
            )
        serializer = getMultiAdapter(
            (OrderData(self.context, uid=self.params[0]), self.request),
            ISerializeToJson,
        )
        return serializer()


class QueryOrdersService(Service):
    """List Orders by given Query for current users vendor

    supported queries (for now, to be extended):
        - state - list with one or more out of:
            cancelled
            finished
            mixed
            new
            processing
            reserved
    """

    def reply(self):
        states = self.request.form.get("state", [])
        if not isinstance(states, list):
            states = [states]
        query = None  # build a repoze catalog query
        if states:
            query = Any("state", states)
        sort_index = self.request.form.get("sort_on", "created")
        reverse = self.request.form.get("sort_order", "") == "reverse"
        manager = OrdersManager(self.context)
        length, orders = manager.orders_data(query, sort_index, reverse)
        result = {
            "length": length,
            "orders": [],
        }
        for order in orders:
            serializer = getMultiAdapter((order, self.request), ISerializeToJson)
            result["orders"].append(serializer())
        return result



# -*- coding: utf-8 -*-
from plone.indexer import indexer
from zope.interface import Interface

import plone.api


@indexer(Interface)
def customer_role(obj):
    """Index users and groups with ``Customer`` role directly on the context.
    Don't index inherited `Customer` role.  Groups are prefixed with ``group:``
    """
    users = obj.users_with_local_role("Customer")  # get non-aquired roles
    ret = [plone.api.group.get(it) and "group:%s" % it or it for it in users]
    return ret



# -*- coding: utf-8 -*-
from bda.plone.orders import message_factory as _
from bda.plone.orders.browser.common import Translate
from bda.plone.orders.contacts import get_contacts_soup
from Products.CMFPlone.resources import add_bundle_on_request
from Products.CMFPlone.utils import safe_unicode
from Products.Five import BrowserView
from repoze.catalog.query import Contains
from repoze.catalog.query import Gt
from yafowil.utils import Tag

import json
import plone.api
import uuid
import yafowil.loader  # noqa: W291,F401


FLOORUID = uuid.UUID(31 * "0" + "1")


class ContactsTable(BrowserView):
    table_id = "bdaplonecontacts"
    data_view_name = "@@contactsdata"

    def __init__(self, context, request):
        super(ContactsTable, self).__init__(context, request)
        add_bundle_on_request(request, "bdajax-jquerytools")
        add_bundle_on_request(request, "bdajax-jquerytools-overlay")
        add_bundle_on_request(request, "datatables")
        add_bundle_on_request(request, "bda-plone-orders")

    def render_get_actions_for_contact(self, colname, record):
        tag = Tag(Translate(self.request))
        site = plone.api.portal.get()
        base_url = site.absolute_url()

        # view orders
        view_orders_target = "%s/@@orders#%s" % (
            base_url,
            str(record.attrs["personal_data.email"]),
        )
        view_orders_attrs = {
            "href": view_orders_target,
            "title": _("view_orders", default=u"View Orders"),
        }
        orders_icon_url = "%s/++resource++bda.plone.orders/orders.png" % base_url
        view_orders_icon = tag("img", src=orders_icon_url)
        view_orders = tag("a", view_orders_icon, **view_orders_attrs)

        # view bookings
        view_bookings_target = "%s/@@bookings#%s" % (
            base_url,
            str(record.attrs["personal_data.email"]),
        )
        view_bookings_attrs = {
            "href": view_bookings_target,
            "title": _("view_bookings", default=u"View Bookings"),
        }
        bookings_icon_url = "%s/++resource++bda.plone.orders/bookings.png" % base_url
        view_bookings_icon = tag("img", src=bookings_icon_url)
        view_bookings = tag("a", view_bookings_icon, **view_bookings_attrs)

        # XXX: view invoices

        return view_orders + view_bookings

    @property
    def ajaxurl(self):
        return "%s/%s" % (self.context.absolute_url(), self.data_view_name)

    @property
    def columns(self):
        return [
            {
                "id": "actions",
                "label": _("actions", default=u"Actions"),
                "renderer": self.render_get_actions_for_contact,
            },
            {"id": "personal_data.email", "label": _("email", default=u"Email")},
            {
                "id": "personal_data.lastname",
                "label": _("lastname", default=u"Last Name"),
            },
            {
                "id": "personal_data.firstname",
                "label": _("firstname", default=u"First Name"),
            },
        ]

    def jsondata(self):
        # json response header needed?
        soup = get_contacts_soup(self.context)
        aaData = list()
        size, result = self.query(soup)

        columns = self.columns
        colnames = [_["id"] for _ in columns]

        def record2list(record, bookings_quantity=None):
            result = list()
            for colname in colnames:
                coldef = self.column_def(colname)
                renderer = coldef.get("renderer")
                if renderer:
                    value = renderer(colname, record)
                else:
                    value = record.attrs.get(colname, "")
                result.append(value)
            return result

        for lazyrecord in self.slice(result):
            aaData.append(record2list(lazyrecord()))

        data = {
            "draw": int(self.request.form["draw"]),
            "recordsTotal": size,
            "recordsFiltered": size,
            "data": aaData,
        }
        self.request.response.setHeader(
            "Content-Type", "application/json; charset=utf-8"
        )
        return json.dumps(data)

    def slice(self, fullresult):
        start = int(self.request.form["start"])
        length = int(self.request.form["length"])
        count = 0
        for lr in fullresult:
            if count >= start and count < (start + length):
                yield lr
            if count >= (start + length):
                break
            count += 1

    def column_def(self, colname):
        for column in self.columns:
            if column["id"] == colname:
                return column

    def _text_checker(self, text):
        # used for query
        if len(text) < 1:
            return None
        return Contains("text", text + "*")

    def query(self, soup):
        # always get all contacts
        query = Gt("uid", FLOORUID)
        req_text = safe_unicode(self.request.get("search[value]", ""))
        text_query = self._text_checker(req_text)
        # take fulltext search into account
        if text_query:
            query = query & text_query

        res = soup.lazy(query, with_size=True)
        size = next(res)
        return size, res



# -*- coding: utf-8 -*-
from bda.plone.cart.restapi.converter import json_compatible_dict
from bda.plone.orders.datamanagers.booking import BookingData
from bda.plone.orders.interfaces.orders import IBookingData
from bda.plone.orders.interfaces.orders import IOrderData
from plone import api
from plone.restapi.interfaces import ISerializeToJson
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.publisher.interfaces import IRequest


@implementer(ISerializeToJson)
@adapter(IBookingData, IRequest)
class SerializeBookingToJson(object):
    def __init__(self, booking_dm, request):
        self.booking_dm = booking_dm
        self.request = request

    def __call__(self):
        data = dict(self.booking_dm.booking.attrs)
        try:
            obj = api.content.get(UID=self.booking_dm.booking.attrs["buyable_uid"],)
            data["path"] = "/".join(obj.getPhysicalPath())
        except ValueError:
            data["path"] = None
        return json_compatible_dict(data)


@implementer(ISerializeToJson)
@adapter(IOrderData, IRequest)
class SerializeOrderToJson(object):
    def __init__(self, order_dm, request):
        self.order_dm = order_dm
        self.request = request

    def __call__(self):
        result = {
            "order": json_compatible_dict(dict(self.order_dm.order.attrs)),
            "bookings": [],
        }
        for booking_record in self.order_dm.bookings:
            serializer = getMultiAdapter(
                (
                    BookingData(
                        self.order_dm.context, booking=booking_record, vendor_uids=[]
                    ),
                    self.request,
                ),
                ISerializeToJson,
            )
            result["bookings"].append(serializer())
        return result



# -*- coding: utf-8 -*-
from repoze.catalog.catalog import Catalog
from repoze.catalog.indexes.field import CatalogFieldIndex
from repoze.catalog.indexes.keyword import CatalogKeywordIndex
from repoze.catalog.indexes.text import CatalogTextIndex
from souper.interfaces import ICatalogFactory
from souper.soup import NodeAttributeIndexer
from souper.soup import NodeTextIndexer
from zope.interface import implementer


@implementer(ICatalogFactory)
class BookingsCatalogFactory(object):
    def __call__(self, context=None):
        catalog = Catalog()
        email_indexer = NodeAttributeIndexer("email")
        catalog[u"email"] = CatalogFieldIndex(email_indexer)
        uid_indexer = NodeAttributeIndexer("uid")
        catalog[u"uid"] = CatalogFieldIndex(uid_indexer)
        buyable_uid_indexer = NodeAttributeIndexer("buyable_uid")
        catalog[u"buyable_uid"] = CatalogFieldIndex(buyable_uid_indexer)
        order_uid_indexer = NodeAttributeIndexer("order_uid")
        catalog[u"order_uid"] = CatalogFieldIndex(order_uid_indexer)
        vendor_uid_indexer = NodeAttributeIndexer("vendor_uid")
        catalog[u"vendor_uid"] = CatalogFieldIndex(vendor_uid_indexer)
        creator_indexer = NodeAttributeIndexer("creator")
        catalog[u"creator"] = CatalogFieldIndex(creator_indexer)
        created_indexer = NodeAttributeIndexer("created")
        catalog[u"created"] = CatalogFieldIndex(created_indexer)
        exported_indexer = NodeAttributeIndexer("exported")
        catalog[u"exported"] = CatalogFieldIndex(exported_indexer)
        title_indexer = NodeAttributeIndexer("title")
        catalog[u"title"] = CatalogFieldIndex(title_indexer)
        state_indexer = NodeAttributeIndexer("state")
        catalog[u"state"] = CatalogFieldIndex(state_indexer)
        salaried_indexer = NodeAttributeIndexer("salaried")
        catalog[u"salaried"] = CatalogFieldIndex(salaried_indexer)
        search_attributes = ["email", "title"]
        text_indexer = NodeTextIndexer(search_attributes)
        catalog[u"text"] = CatalogTextIndex(text_indexer)
        return catalog


@implementer(ICatalogFactory)
class OrdersCatalogFactory(object):
    def __call__(self, context=None):
        catalog = Catalog()
        uid_indexer = NodeAttributeIndexer("uid")
        catalog[u"uid"] = CatalogFieldIndex(uid_indexer)
        email_indexer = NodeAttributeIndexer("personal_data.email")
        catalog[u"personal_data.email"] = CatalogFieldIndex(email_indexer)
        ordernumber_indexer = NodeAttributeIndexer("ordernumber")
        catalog[u"ordernumber"] = CatalogFieldIndex(ordernumber_indexer)
        booking_uids_indexer = NodeAttributeIndexer("booking_uids")
        catalog[u"booking_uids"] = CatalogKeywordIndex(booking_uids_indexer)
        vendor_uids_indexer = NodeAttributeIndexer("vendor_uids")
        buyable_uids_indexer = NodeAttributeIndexer("buyable_uids")
        catalog[u"buyable_uids"] = CatalogKeywordIndex(buyable_uids_indexer)
        catalog[u"vendor_uids"] = CatalogKeywordIndex(vendor_uids_indexer)
        creator_indexer = NodeAttributeIndexer("creator")
        catalog[u"creator"] = CatalogFieldIndex(creator_indexer)
        created_indexer = NodeAttributeIndexer("created")
        catalog[u"created"] = CatalogFieldIndex(created_indexer)
        firstname_indexer = NodeAttributeIndexer("personal_data.firstname")
        catalog[u"personal_data.firstname"] = CatalogFieldIndex(firstname_indexer)
        lastname_indexer = NodeAttributeIndexer("personal_data.lastname")
        catalog[u"personal_data.lastname"] = CatalogFieldIndex(lastname_indexer)
        city_indexer = NodeAttributeIndexer("billing_address.city")
        catalog[u"billing_address.city"] = CatalogFieldIndex(city_indexer)
        search_attributes = [
            "personal_data.lastname",
            "personal_data.firstname",
            "personal_data.email",
            "billing_address.city",
            "ordernumber",
        ]
        text_indexer = NodeTextIndexer(search_attributes)
        catalog[u"text"] = CatalogTextIndex(text_indexer)
        # state on order only used for sorting in orders table
        state_indexer = NodeAttributeIndexer("state")
        catalog[u"state"] = CatalogFieldIndex(state_indexer)
        # salaried on order only used for sorting in orders table
        salaried_indexer = NodeAttributeIndexer("salaried")
        catalog[u"salaried"] = CatalogFieldIndex(salaried_indexer)
        return catalog



# -*- coding: utf-8 -*-
from Products.CMFCore.permissions import setDefaultRoles


# view individual order
ViewOrderDirectly = "bda.plone.orders: View Order Directly"
setDefaultRoles(ViewOrderDirectly, ("Manager", "Site Administrator", "Authenticated"))

# view own orders
ViewOwnOrders = "bda.plone.orders: View Own Orders"
setDefaultRoles(ViewOwnOrders, ("Manager", "Site Administrator", "Authenticated"))

# view all orders
ViewOrders = "bda.plone.orders: View Orders"
setDefaultRoles(ViewOrders, ("Manager", "Site Administrator", "Authenticated"))

# modify orders
ModifyOrders = "bda.plone.orders: Modify Orders"
setDefaultRoles(ModifyOrders, ("Manager", "Site Administrator", "Vendor"))

# export orders
ExportOrders = "bda.plone.orders: Export Orders"
setDefaultRoles(ExportOrders, ("Manager", "Site Administrator", "Vendor"))

# manage templates
ManageTemplates = "bda.plone.orders: Manage Templates"
setDefaultRoles(ManageTemplates, ("Manager", "Site Administrator", "Vendor"))

# delegate customer role
DelegateCustomerRole = "bda.plone.orders: Delegate Customer Role"
setDefaultRoles(DelegateCustomerRole, ("Manager", "Site Administrator"))

# delegate vendor role
DelegateVendorRole = "bda.plone.orders: Delegate Vendor Role"
setDefaultRoles(DelegateVendorRole, tuple())



# -*- coding: utf-8 -*-
from AccessControl import Unauthorized
from Acquisition import aq_parent
from bda.plone.cart.cartitem import get_item_stock
from bda.plone.cart.utils import get_object_by_uid
from bda.plone.orders import message_factory as _
from bda.plone.orders import permissions
from bda.plone.orders import safe_encode
from bda.plone.orders import safe_filename
from bda.plone.orders.browser.common import customers_form_vocab
from bda.plone.orders.browser.common import vendors_form_vocab
from bda.plone.orders.common import DT_FORMAT
from bda.plone.orders.common import get_bookings_soup
from bda.plone.orders.common import get_order
from bda.plone.orders.common import get_orders_soup
from bda.plone.orders.common import get_vendor_uids_for
from bda.plone.orders.common import get_vendors_for
from bda.plone.orders.contacts import get_contacts_soup
from bda.plone.orders.datamanagers.order import OrderData
from bda.plone.orders.interfaces import IBuyable
from decimal import Decimal
from io import BytesIO
from odict import odict
from plone.uuid.interfaces import IUUID
from Products.CMFPlone.utils import getToolByName
from Products.CMFPlone.utils import safe_unicode
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from repoze.catalog.query import Any
from repoze.catalog.query import Eq
from repoze.catalog.query import Ge
from repoze.catalog.query import Le
from yafowil.base import ExtractionError
from yafowil.controller import Controller
from yafowil.plone.form import YAMLForm

import csv
import csv23
import datetime
import plone.api
import uuid
import yafowil.loader  # noqa


class DialectExcelWithColons(csv.excel):
    delimiter = ";"


csv23.register_dialect("excel-colon", DialectExcelWithColons)

EXPORT_DT_FORMAT = DT_FORMAT

EXPORT_CHARSET = "UTF-8"

ORDER_EXPORT_ATTRS = [
    "uid",
    "created",
    "ordernumber",
    "cart_discount_net",
    "cart_discount_vat",
    "personal_data.company",
    "personal_data.email",
    "personal_data.gender",
    "personal_data.firstname",
    "personal_data.phone",
    "personal_data.lastname",
    "billing_address.city",
    "billing_address.country",
    "billing_address.street",
    "billing_address.zip",
    "delivery_address.alternative_delivery",
    "delivery_address.city",
    "delivery_address.company",
    "delivery_address.country",
    "delivery_address.firstname",
    "delivery_address.street",
    "delivery_address.lastname",
    "delivery_address.zip",
    "order_comment.comment",
    "payment_selection.payment",
]
COMPUTED_ORDER_EXPORT_ATTRS = odict()
CONTACT_EXPORT_ATTRS = [
    "cid",
]
COMPUTED_CONTACT_EXPORT_ATTRS = odict()
BOOKING_EXPORT_ATTRS = [
    "title",
    "buyable_comment",
    "buyable_count",
    "quantity_unit",
    "net",
    "discount_net",
    "vat",
    "currency",
    "state",
    "salaried",
    "exported",
]
COMPUTED_BOOKING_EXPORT_ATTRS = odict()


def buyable_available(context, booking):
    obj = get_object_by_uid(context, booking.attrs["buyable_uid"])
    if not obj:
        return None
    item_stock = get_item_stock(obj)
    if not item_stock:
        return None
    return item_stock.available


def buyable_overbook(context, booking):
    obj = get_object_by_uid(context, booking.attrs["buyable_uid"])
    if not obj:
        return None
    item_stock = get_item_stock(obj)
    if not item_stock:
        return None
    return item_stock.overbook


def buyable_url(context, booking):
    obj = get_object_by_uid(context, booking.attrs["buyable_uid"])
    if not obj:
        return None
    return obj.absolute_url()


COMPUTED_BOOKING_EXPORT_ATTRS["buyable_available"] = buyable_available
COMPUTED_BOOKING_EXPORT_ATTRS["buyable_overbook"] = buyable_overbook
COMPUTED_BOOKING_EXPORT_ATTRS["buyable_url"] = buyable_url


def cleanup_for_csv(value):
    """Cleanup a value for CSV export.
    """
    if isinstance(value, datetime.datetime):
        value = value.strftime(EXPORT_DT_FORMAT)
    if value == "-":
        value = ""
    return value


class ExportMixin(object):

    def export_val(self, record, attr_name):
        """Get attribute from record and cleanup.
        Since the record object is available, you can return aggregated values.
        """
        val = record.attrs.get(attr_name)
        return cleanup_for_csv(val)

    def csv(self):
        # get orders soup
        orders_soup = get_orders_soup(self.context)
        # get bookings soup
        bookings_soup = get_bookings_soup(self.context)
        # get contacts soup
        contacts_soup = get_contacts_soup(self.context)
        # fetch user vendor uids
        vendor_uids = get_vendor_uids_for()
        # base query for time range
        query = Ge("created", self.from_date) & Le("created", self.to_date)
        # filter by given vendor uid or user vendor uids
        vendor_uid = self.vendor
        if vendor_uid:
            vendor_uid = uuid.UUID(vendor_uid)
            # raise if given vendor uid not in user vendor uids
            if vendor_uid not in vendor_uids:
                raise Unauthorized
            query = query & Any("vendor_uids", [vendor_uid])
        else:
            query = query & Any("vendor_uids", vendor_uids)
        # filter by customer if given
        customer = self.customer
        if customer:
            query = query & Eq("creator", customer)
        # prepare csv writer
        bio = BytesIO()
        ex = csv23.writer(bio, dialect="excel-colon", encoding=EXPORT_CHARSET)
        # exported column keys as first line
        ex.writerow(
            ORDER_EXPORT_ATTRS
            + list(COMPUTED_ORDER_EXPORT_ATTRS.keys())
            + CONTACT_EXPORT_ATTRS
            + list(COMPUTED_CONTACT_EXPORT_ATTRS.keys())
            + BOOKING_EXPORT_ATTRS
            + list(COMPUTED_BOOKING_EXPORT_ATTRS.keys())
        )
        # query orders
        for order in orders_soup.query(query):
            # restrict order bookings for current vendor_uids
            order_data = OrderData(self.context, order=order, vendor_uids=vendor_uids)
            order_attrs = list()
            # order export attrs
            for attr_name in ORDER_EXPORT_ATTRS:
                val = self.export_val(order, attr_name)
                order_attrs.append(val)
            # computed order export attrs
            for attr_name in COMPUTED_ORDER_EXPORT_ATTRS:
                cb = COMPUTED_ORDER_EXPORT_ATTRS[attr_name]
                val = cb(self.context, order_data)
                val = cleanup_for_csv(val)
                order_attrs.append(val)
            if CONTACT_EXPORT_ATTRS:
                contact_attrs = list()
                contact = list(
                    contacts_soup.query(Eq("uid", order.attrs["contact_uid"]))
                )
                if contact:
                    # contact export attrs
                    for attr_name in CONTACT_EXPORT_ATTRS:
                        val = self.export_val(contact[0], attr_name)
                        contact_attrs.append(val)
                    # computed contact export attrs
                    for attr_name in COMPUTED_CONTACT_EXPORT_ATTRS:
                        cb = COMPUTED_CONTACT_EXPORT_ATTRS[attr_name]
                        val = cb(self.context, contact[0])
                        val = cleanup_for_csv(val)
                        contact_attrs.append(val)

            for booking in order_data.bookings:
                booking_attrs = list()
                # booking export attrs
                for attr_name in BOOKING_EXPORT_ATTRS:
                    val = self.export_val(booking, attr_name)
                    booking_attrs.append(val)
                # computed booking export attrs
                for attr_name in COMPUTED_BOOKING_EXPORT_ATTRS:
                    cb = COMPUTED_BOOKING_EXPORT_ATTRS[attr_name]
                    val = cb(self.context, booking)
                    val = cleanup_for_csv(val)
                    booking_attrs.append(val)
                ex.writerow(order_attrs + contact_attrs + booking_attrs)
                booking.attrs["exported"] = True
                bookings_soup.reindex(booking)
        ret = bio.getvalue()
        bio.close()
        return ret


class ExportOrdersForm(YAMLForm, ExportMixin, BrowserView):
    browser_template = ViewPageTemplateFile("templates/export.pt")
    form_template = "bda.plone.orders.browser:forms/orders_export.yaml"
    message_factory = _
    action_resource = "exportorders"

    def __call__(self):
        # check if authenticated user is vendor
        if not get_vendors_for():
            raise Unauthorized
        self.prepare()
        controller = Controller(self.form, self.request)
        if not controller.next:
            self.rendered_form = controller.rendered
            return self.browser_template(self)
        return controller.next

    def vendor_vocabulary(self):
        return vendors_form_vocab()

    @property
    def vendor_mode(self):
        return len(vendors_form_vocab()) > 2 and "edit" or "skip"

    def customer_vocabulary(self):
        return customers_form_vocab()

    @property
    def customer_mode(self):
        return len(customers_form_vocab()) > 2 and "edit" or "skip"

    def from_before_to(self, widget, data):
        from_date = data.fetch("exportorders.from").extracted
        to_date = data.fetch("exportorders.to").extracted
        if to_date <= from_date:
            raise ExtractionError(
                _("from_date_before_to_date", default=u"From-date after to-date")
            )
        return to_date

    def export(self, widget, data):
        self.vendor = self.request.form.get("exportorders.vendor")
        self.customer = self.request.form.get("exportorders.customer")
        self.from_date = data.fetch("exportorders.from").extracted
        self.to_date = data.fetch("exportorders.to").extracted

    def csv(self, request):
        ret = super(ExportOrdersForm, self).csv()
        # create and return response
        s_start = self.from_date.strftime("%G-%m-%d_%H-%M-%S")
        s_end = self.to_date.strftime("%G-%m-%d_%H-%M-%S")
        filename = "orders-export-%s-%s.csv" % (s_start, s_end)
        self.request.response.setHeader("Content-Type", "text/csv")
        self.request.response.setHeader(
            "Content-Disposition", "attachment; filename=%s" % filename
        )
        return ret


class ExportOrdersContextual(BrowserView):
    def __call__(self):
        user = plone.api.user.get_current()
        # check if authenticated user is vendor
        if not user.checkPermission(permissions.ModifyOrders, self.context):
            raise Unauthorized

        # Special case for constructed objects like IEventOccurrence from
        # plone.app.event
        title = self.context.title or aq_parent(self.context).title

        filename = u"{0}_{1}.csv".format(
            safe_unicode(title),
            safe_unicode(datetime.datetime.now().strftime("%Y-%m-%d_%H-%M")),
        )
        filename = safe_filename(filename)
        resp = self.request.response
        resp.setHeader("content-type", "text/csv; charset={0}".format(EXPORT_CHARSET))
        resp.setHeader("content-disposition", "attachment;filename={}".format(filename))
        return self.get_csv()

    def export_val(self, record, attr_name):
        """Get attribute from record and cleanup.
        Since the record object is available, you can return aggregated values.
        """
        val = record.attrs.get(attr_name)
        return cleanup_for_csv(val)

    def get_csv(self):
        context = self.context

        # prepare csv writer
        bio = BytesIO()
        ex = csv23.writer(bio, dialect="excel-colon", encoding=EXPORT_CHARSET)
        # exported column keys as first line
        ex.writerow(
            ORDER_EXPORT_ATTRS
            + list(COMPUTED_ORDER_EXPORT_ATTRS.keys())
            + BOOKING_EXPORT_ATTRS
            + list(COMPUTED_BOOKING_EXPORT_ATTRS.keys())
        )

        bookings_soup = get_bookings_soup(context)

        # First, filter by allowed vendor areas
        vendor_uids = get_vendor_uids_for()
        query_b = Any("vendor_uid", vendor_uids)

        # Second, query for the buyable
        query_cat = {}
        query_cat["object_provides"] = IBuyable.__identifier__
        query_cat["path"] = "/".join(context.getPhysicalPath())
        cat = getToolByName(context, "portal_catalog")
        res = cat(**query_cat)
        buyable_uids = [IUUID(it.getObject()) for it in res]

        query_b = query_b & Any("buyable_uid", buyable_uids)

        all_orders = {}
        for booking in bookings_soup.query(query_b):
            booking_attrs = []
            # booking export attrs
            for attr_name in BOOKING_EXPORT_ATTRS:
                val = self.export_val(booking, attr_name)
                booking_attrs.append(val)
            # computed booking export attrs
            for attr_name in COMPUTED_BOOKING_EXPORT_ATTRS:
                cb = COMPUTED_BOOKING_EXPORT_ATTRS[attr_name]
                val = cb(context, booking)
                val = cleanup_for_csv(val)
                booking_attrs.append(val)

            # create order_attrs, if it doesn't exist
            order_uid = booking.attrs.get("order_uid")
            if order_uid not in all_orders:
                order = get_order(context, order_uid)
                order_data = OrderData(context, order=order, vendor_uids=vendor_uids)
                order_attrs = []
                # order export attrs
                for attr_name in ORDER_EXPORT_ATTRS:
                    val = self.export_val(order, attr_name)
                    order_attrs.append(val)
                # computed order export attrs
                for attr_name in COMPUTED_ORDER_EXPORT_ATTRS:
                    cb = COMPUTED_ORDER_EXPORT_ATTRS[attr_name]
                    val = cb(self.context, order_data)
                    val = cleanup_for_csv(val)
                    order_attrs.append(val)
                all_orders[order_uid] = order_attrs

            ex.writerow(all_orders[order_uid] + booking_attrs)

            # TODO: also set for contextual exports? i'd say no.
            # booking.attrs['exported'] = True
            # bookings_soup.reindex(booking)

        ret = bio.getvalue()
        bio.close()
        return ret



# -*- coding: utf-8 -*-
from Acquisition import aq_inner
from Acquisition import aq_parent
from bda.plone.cart.utils import get_object_by_uid
from bda.plone.orders import permissions
from bda.plone.orders.interfaces.markers import IVendor
from plone.uuid.interfaces import IUUID
from Products.CMFPlone.interfaces import IPloneSiteRoot
from repoze.catalog.query import Any
from repoze.catalog.query import Eq
from souper.soup import get_soup
from zope.deferredimport import deprecated

import logging
import plone.api
import uuid


deprecated(
    "Import from new location 'catalogfactories' instead",
    BookingsCatalogFactory="bda.plone.orders.catalogfactories:BookingsCatalogFactory",
    OrdersCatalogFactory="bda.plone.orders.catalogfactories:OrdersCatalogFactory",
)
deprecated(
    "Import from new location 'checkout' instead",
    OrderCheckoutAdapter="bda.plone.orders.checkout:OrderCheckoutAdapter",
)
deprecated(
    "Import from new location 'datamanagers.base' instead",
    OrderState="bda.plone.orders.datamanagers.base:BaseState",
    calculate_order_state="bda.plone.orders.datamanagers.base:calculate_order_state",
    calculate_order_salaried="bda.plone.orders.datamanagers.base:calculate_order_salaried",
)
deprecated(
    "Import from new location 'datamanagers.order' instead",
    create_order_number="bda.plone.orders.datamanagers.order:create_order_number",
    OrderData="bda.plone.orders.datamanagers.order:OrderData",
)
deprecated(
    "Import from new location 'datamanagers.booking' instead",
    BookingData="bda.plone.orders.datamanagers.booking:BookingData",
    booking_update_comment="bda.plone.orders.datamanagers.booking:booking_update_comment",
)
deprecated(
    "Import from new location 'datamanagers.buyable' instead",
    BuyableData="bda.plone.orders.datamanagers.buyable:BuyableData",
)
deprecated(
    "Import from new location 'datamanagers.payment' instead",
    PaymentData="bda.plone.orders.datamanagers.payment:PaymentData",
    payment_success="bda.plone.orders.datamanagers.payment:payment_success",
    payment_failed="bda.plone.orders.datamanagers.payment:payment_failed",
)

logger = logging.getLogger(__name__)


DT_FORMAT = "%d.%m.%Y %H:%M"
DT_FORMAT_SHORT = "%d.%m.%Y"


def get_orders_soup(context):
    return get_soup("bda_plone_orders_orders", context)


def get_bookings_soup(context):
    return get_soup("bda_plone_orders_bookings", context)


def get_order(context, uid):
    if not isinstance(uid, uuid.UUID):
        uid = uuid.UUID(uid)
    soup = get_orders_soup(context)
    return [it for it in soup.query(Eq("uid", uid))][0]


def acquire_vendor_or_shop_root(context):
    """Returns the acquired vendor or the main shop by traversing up the
    content tree, starting from a context.

    :param context: The context to start searching for the nearest vendor.
    :type context: Content object
    :returns: The vendor, a shop item is belonging to.
    :rtype: Content object
    """
    if not context:
        message = u"No context given to acquire vendor or shop root from"
        raise ValueError(message)
    while not IVendor.providedBy(context) and not IPloneSiteRoot.providedBy(context):
        context = aq_parent(aq_inner(context))
    return context


def get_all_vendors():
    """Get all available vendor areas.

    :returns: Vendor area enabled content objects.
    :rtype: List of content objects.
    """
    cat = plone.api.portal.get_tool("portal_catalog")
    query = {"object_provides": IVendor.__identifier__}
    vendors = [brain.getObject() for brain in cat(**query)]
    return vendors + [plone.api.portal.get()]


def get_vendor_by_uid(context, vendor_uid):
    """Return vendor object by uid or site root as fallback if requested vendor
    not exists.

    :param vendor_uid: Vendor uid.
    :type vendor_uid: string or uuid.UUID object
    :returns: Vendor Object
    :rtype: IVendor implementing content Object
    """
    if not isinstance(vendor_uid, uuid.UUID):
        vendor_uid = uuid.UUID(vendor_uid)
    vendor = get_object_by_uid(context, vendor_uid)
    if vendor is None:
        vendor = plone.api.portal.get()
    return vendor


def get_vendors_for(user=None):
    """Gel all vendor containers a given or authenticated user has vendor
    permissions for.

    :param user: Optional user object to check permissions on vendor areas. If
                 no user object is give, the current user is used.
    :type user: MemberData object
    :returns: Allowed vendor area enabled content objects.
    :rtype: List of content objects.
    """
    if user is None:
        user = plone.api.user.get_current()

    def permitted(obj):
        return plone.api.user.has_permission(permissions.ModifyOrders, obj=obj)

    return [vendor for vendor in get_all_vendors() if permitted(vendor)]


def get_vendor_uids_for(user=None):
    """Gel all vendor container uids a given or authenticated user has vendor
    permissions for.

    :param user: Optional user object to check permissions on vendor areas. If
                 no user object is give, the current user is used.
    :type user: MemberData object
    :returns: Allowed vendor area uids for given or authenticated member.
    :rtype: List of uuid.UUID objects.
    """
    return [uuid.UUID(IUUID(vendor)) for vendor in get_vendors_for(user=user)]


# TODO: used in vocabularies, remove if possible
def get_vendor_order_uids_for(context, user=None):
    """Get order uids all orders a given or current user has vendor
    permissions for.

    :param user: Optional user object to check permissions on vendor areas. If
                 no user object is give, the current user is used.
    :type user: MemberData object
    :returns: List of order UUID objects.
    :rtype: List of uuid.UUID
    """
    vendors = get_vendor_uids_for(user=user)
    soup = get_bookings_soup(context)
    res = soup.query(Any("vendor_uid", vendors))
    order_uids = {booking.attrs["order_uid"] for booking in res}
    return order_uids


# Patch this tuple with booking states which should be ignored at initial
# billing.
# WARNING: see notes in ``is_billable_booking`` docs.
BOOKING_BILLABLE_IGNORE_STATES = tuple()


def is_billable_booking(booking):
    """Return True, if booking is billable and should be included in order
    summary calculations.

    To be used in Pythons filter function:

        filter(is_billable_booking, bookings)

    WARNING:

    This function has been added for bda.plone.ticketshop to skip billing
    sold out tickets. When selling tickets, a backorder is handled that
    someone waits until a ticket order from another customer gets cancelled.

    But this behavior does not apply to most shop implementations where goods
    are sold. Here a backorder normally means a temporary unavailability of
    buyable items and has effect to delivery time, not billing.

    Since bda.plone.orders does not properly implement the full UI to handle
    partly billed bookings (carry back unbilled backorders, incomplete order
    and invoive views), this function is treated as hack, but is not removed
    due to the needs of ticketshop unless the related usecase is properly
    implemented.

    See https://github.com/bluedynamics/bda.plone.orders/issues/45
    """
    return booking.attrs["state"] not in BOOKING_BILLABLE_IGNORE_STATES



# -*- coding: utf-8 -*-
from Products.CMFPlone.utils import safe_unicode
from zope.i18nmessageid import MessageFactory

import gettext
import pycountry
import six
import unicodedata


message_factory = MessageFactory("bda.plone.orders")


def safe_encode(string):
    """Safely unicode objects to UTF-8. If it's a binary string, just return
    it.
    """
    if isinstance(string, six.string_types):
        return safe_unicode(string).encode("utf-8")
    return string


def safe_filename(value):
    """
    Convert to ASCII if 'allow_unicode' is False. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace.

    Ideas from:
    https://github.com/django/django/blob/master/django/utils/text.py
    """
    value = safe_unicode(value)
    value = unicodedata.normalize("NFKD", value).encode("ascii", "ignore")
    return value.replace(" ", "-").strip()


def get_country_name(country_code, lang="en"):
    """Return the translated country name for a given pycountry country code.
    """

    country_name = pycountry.countries.get(numeric=country_code).name
    trans = gettext.translation("iso3166-1", pycountry.LOCALES_DIR, languages=["de"])
    return safe_unicode(trans.gettext(country_name))



# -*- coding: utf-8 -*-
from Acquisition import aq_parent
from bda.plone.ajax import ajax_continue
from bda.plone.ajax import ajax_form_fiddle
from bda.plone.ajax import AjaxMessage
from bda.plone.ajax import AjaxOverlay
from bda.plone.orders import message_factory as _
from bda.plone.orders.browser.mailtemplates import TEMPLATE
from bda.plone.orders.datamanagers.order import OrderData
from bda.plone.orders.interfaces import IDynamicMailTemplateLibrary
from bda.plone.orders.mailnotify import MailNotify
from bda.plone.orders.vocabularies import salaried_vocab
from bda.plone.orders.vocabularies import state_vocab
from node.utils import UNSET
from Products.Five import BrowserView
from yafowil.base import ExtractionError
from yafowil.plone.form import YAMLBaseForm
from zope.component.interfaces import ISite
from zope.i18n import translate

import json


def mail_gender_vocab():
    return {
        "-": _("mister_misses", default=u"Mister/Misses"),
        "male": _("mister", default=u"Mister"),
        "female": _("misses", default=u"Misses"),
    }


class NotifyCustomers(YAMLBaseForm):
    """Notify customers form.
    """

    form_template = "bda.plone.orders.browser:forms/notify_customers.yaml"
    message_factory = _

    def __init__(self, context, request):
        # Find a ISite context when called anywhere.
        # The context must be a ISite, but the view can be called anywhere.
        def _acquire_site(context):
            if ISite.providedBy(context):
                return context
            return _acquire_site(aq_parent(context))

        self.context = _acquire_site(context)
        self.request = request

    def form_action(self, widget, data):
        return "%s/ajaxform?form_name=notify_customers" % self.context.absolute_url()

    def template_value(self, widget, data):
        if data.extracted and data.extracted["template"] != "-":
            tpllib = IDynamicMailTemplateLibrary(self.context)
            try:
                return tpllib[data.extracted["template"]]
            except KeyError:
                pass
        return UNSET

    def template_vocabulary(self, widget, data):
        vocab = [("-", _("no_template_selected", default=u"No template selected"))]
        tpllib = IDynamicMailTemplateLibrary(self.context)
        for key in tpllib.keys():
            vocab.append((key, key))
        return vocab

    def text_value(self, widget, data):
        return UNSET

    def validate_tpl(self, widget, data):
        if not data.extracted:
            return data.extracted
        state, msg = TEMPLATE.validate(data.extracted)
        if not state:
            raise ExtractionError(msg)
        return data.extracted

    def _sendmail(self, notifier, uid, tpl, subject):
        order_data = OrderData(self.context, uid=uid)
        order = order_data.order
        data = {}
        for key in TEMPLATE.defaults:
            # XXX: refactor and use configurable callbacks
            # special case gender
            if key == "personal_data.gender":
                data[key] = translate(
                    mail_gender_vocab()[order.attrs[key]], context=self.request
                )
                continue
            # special case salaried
            if key == "salaried":
                data[key] = translate(
                    salaried_vocab()[order_data.salaried], context=self.request
                )
                continue
            # special case state
            if key == "state":
                data[key] = translate(
                    state_vocab()[order_data.state], context=self.request
                )
                continue
            # read attrs from order
            if key in order.attrs:
                data[key] = order.attrs[key]
        body = TEMPLATE(tpl, data).encode("utf-8")
        notifier.send(subject, order.attrs["personal_data.email"], body)

    def send(self, widget, data):
        tpl = data["text"].extracted
        subject = data["subject"].extracted
        notifier = MailNotify(self.context)
        for uid in self.request.form.get("uids", []):
            if not uid:
                continue
            self._sendmail(notifier, uid, tpl, subject)

    def ajax_url(self, widget, data):
        url = u"{0}/@@load_notification_template".format(self.context.absolute_url())
        return {"tplurl": url}

    def send_success(self, request):
        message = translate(
            _("customers_notified_success", default=u"Mail to customers sent"),
            context=self.request,
        )
        continuation = [AjaxOverlay(close=True), AjaxMessage(message, "info", None)]
        ajax_continue(self.request, continuation)
        return True

    def __call__(self):
        # XXX: security check -> current user has valid vandor area?
        ajax_form_fiddle(self.request, "form[id=form-notify_customers]", "replace")
        return self.render_form()


class LoadTemplate(BrowserView):
    def __call__(self):
        # XXX: security check -> current user has valid vandor area?
        self.request.response.setHeader("Content-Type", "application/json")
        tpllib = IDynamicMailTemplateLibrary(self.context)
        try:
            tpl = tpllib[self.request.form["name"]]
        except Exception:
            tpl = ""
        return json.dumps({"tpl": tpl})



# -*- coding: utf-8 -*-
from bda.plone.orders.common import get_order
from bda.plone.orders.common import is_billable_booking
from bda.plone.orders.datamanagers.base import BaseState
from bda.plone.orders.datamanagers.base import calculate_order_salaried
from bda.plone.orders.datamanagers.base import calculate_order_state
from bda.plone.orders.interfaces.orders import IOrderData
from decimal import Decimal
from repoze.catalog.query import Any
from repoze.catalog.query import Eq
from six.moves import filter
from zope.interface import implementer

import time
import uuid


def create_ordernumber():
    onum = hash(time.time())
    if onum < 0:
        return "0{0:d}".format(abs(onum))
    return "1{0:d}".format(onum)


@implementer(IOrderData)
class OrderData(BaseState):
    """Object for extracting order information.
    """

    def __init__(self, context, uid=None, order=None, vendor_uids=[]):
        """Create order data object by criteria

        :param context: Context to work with
        :type object: Plone or Content instance
        :param uid: Order uid. XOR with order
        :type uid: string or uuid.UUID object
        :param order: Order record. XOR with uid
        :type order: souper.soup.Record object
        :param vendor_uids: Vendor uids, used to filter order bookings.
        :type vendor_uids: List of vendor uids as string or uuid.UUID object.
        """
        if not (bool(uid) != bool(order)):  # ^= xor
            raise ValueError("Parameters 'uid' and 'order' are mutually exclusive.")
        if uid and not isinstance(uid, uuid.UUID):
            uid = uuid.UUID(uid)
        vendor_uids = [uuid.UUID(str(vuid)) for vuid in vendor_uids]
        self.context = context
        self._uid = uid
        self._order = order
        self.vendor_uids = vendor_uids

    @property
    def uid(self):
        if self._uid:
            return self._uid
        return self.order.attrs["uid"]

    @property
    def order(self):
        if self._order:
            return self._order
        return get_order(self.context, self.uid)

    @property
    def bookings(self):
        soup = self.bookings_soup
        query = Eq("order_uid", self.uid)
        if self.vendor_uids:
            query = query & Any("vendor_uid", self.vendor_uids)
        return soup.query(query)

    @property
    def currency(self):
        ret = None
        for booking in self.bookings:
            val = booking.attrs["currency"]
            if ret and ret != val:
                msg = (
                    u"Order contains bookings with inconsistent "
                    u"currencies {0} != {1}".format(ret, val)
                )
                raise ValueError(msg)
            ret = val
        return ret

    @property
    def state(self):
        return calculate_order_state(self.bookings)

    @state.setter
    def state(self, value):
        # XXX: currently we need to delete attributes before setting to a new
        #      value in order to persist change. fix in appropriate place.
        bookings = self.bookings
        for booking in bookings:
            self.update_item_stock(booking, booking.attrs["state"], value)
            del booking.attrs["state"]
            booking.attrs["state"] = value
        order = self.order
        del order.attrs["state"]
        order.attrs["state"] = value
        self.reindex_bookings(bookings)
        self.reindex_order(order)

    @property
    def salaried(self):
        return calculate_order_salaried(self.bookings)

    @salaried.setter
    def salaried(self, value):
        # XXX: currently we need to delete attributes before setting to a new
        #      value in order to persist change. fix in appropriate place.
        bookings = self.bookings
        for booking in bookings:
            del booking.attrs["salaried"]
            booking.attrs["salaried"] = value
        order = self.order
        del order.attrs["salaried"]
        order.attrs["salaried"] = value
        self.reindex_bookings(bookings)
        self.reindex_order(order)

    @property
    def tid(self):
        ret = set()
        for booking in self.bookings:
            tid = booking.attrs.get("tid", None)
            if tid:
                ret.add(tid)
        return ret

    @tid.setter
    def tid(self, value):
        for booking in self.bookings:
            booking.attrs["tid"] = value

    @property
    def net(self):
        ret = Decimal(0)
        for booking in filter(is_billable_booking, self.bookings):
            count = Decimal(booking.attrs["buyable_count"])
            net = Decimal(booking.attrs.get("net", 0))
            discount_net = Decimal(booking.attrs["discount_net"])
            ret += (net - discount_net) * count
        return ret

    @property
    def vat(self):
        ret = Decimal(0)
        for booking in filter(is_billable_booking, self.bookings):
            count = Decimal(booking.attrs["buyable_count"])
            net = Decimal(booking.attrs.get("net", 0))
            discount_net = Decimal(booking.attrs["discount_net"])
            item_net = net - discount_net
            ret += (item_net * Decimal(booking.attrs.get("vat", 0)) / 100) * count
        return ret

    @property
    def discount_net(self):
        return Decimal(self.order.attrs["cart_discount_net"])

    @property
    def discount_vat(self):
        return Decimal(self.order.attrs["cart_discount_vat"])

    @property
    def shipping_net(self):
        return Decimal(self.order.attrs["shipping_net"])

    @property
    def shipping_vat(self):
        return Decimal(self.order.attrs["shipping_vat"])

    @property
    def shipping(self):
        return Decimal(self.order.attrs["shipping"])

    @property
    def total(self):
        # XXX: use decimal
        total = self.net - self.discount_net + self.vat - self.discount_vat
        return total + self.shipping

    def remove(self):
        """removes this order from the database.
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
from bda.plone.orders import interfaces as ifaces
from bda.plone.orders import message_factory as _
from bda.plone.orders.datamanagers.order import OrderData
from Products.Five import BrowserView


###############################################################################
# views
###############################################################################


class OrderDone(BrowserView):
    """Landing page after order has been done.
    """

    # XXX: provide different headings and texts for states reservation and
    #      mixed
    reservation_states = (ifaces.STATE_RESERVED, ifaces.STATE_MIXED)

    @property
    def order_data(self):
        return OrderData(self.context, uid=self.request.get("uid"))

    @property
    def heading(self):
        try:
            if self.order_data.state in self.reservation_states:
                return _("reservation_done", default=u"Reservation Done")
            return _("order_done", default=u"Order Done")
        except ValueError:
            return _("unknown_order", default=u"Unknown Order")

    @property
    def id(self):
        try:
            return self.order_data.order.attrs["ordernumber"]
        except ValueError:
            return _("unknown", default=u"Unknown")

    @property
    def text(self):
        try:
            if self.order_data.state in self.reservation_states:
                return _("reservation_text", default=u"Thanks for your Reservation.")
            return _("order_text", default=u"Thanks for your Order.")
        except ValueError:
            return _("unknown_order_text", default=u"Sorry, this order does not exist.")
